import copy, json
import numpy as np
try:
    import torch
    import torch.nn.functional as F
except:
    print('warning! pytorch not installed')
    print('警告! 没有安装pytorch，所有pytorch相关函数不可用!')
    class torch():
        Tensor = Exception
from functools import wraps


class ConfigCache(object):
    def __init__(self) -> None:
        super().__init__()
        self.init = False

    def read_cfg(self):
        from config import GlobalConfig
        if GlobalConfig.cfg_ready:
            self.device_ = GlobalConfig.device
            self.use_float64_ = GlobalConfig.use_float64
            self.init = True

    @property
    def device(self):
        if not self.init: self.read_cfg()
        assert self.init, ('cfg not ready!')
        return self.device_

    @property
    def use_float64(self):
        if not self.init: self.read_cfg()
        assert self.init, ('cfg not ready!')
        return self.use_float64_

cfg = ConfigCache()

def pt_inf():
    # if not cfg.init: cfg.read_cfg()
    pt_dtype = torch.float64 if cfg.use_float64 else torch.float32
    return torch.tensor(np.inf, dtype=pt_dtype, device=cfg.device)

def pt_nan():
    # if not cfg.init: cfg.read_cfg()
    pt_dtype = torch.float64 if cfg.use_float64 else torch.float32
    return torch.tensor(np.nan, dtype=pt_dtype, device=cfg.device)


def vis_mat(mat):
    mat = mat.astype(np.float)
    mat = mat - mat.min()
    mat = mat / mat.max()
    import matplotlib.pyplot as plt
    import matplotlib.image as mpimg

    imgplot = plt.imshow(mat)
    plt.xlabel("cols, 2rd dim")
    plt.ylabel("lines, 1st dim")
    plt.show()


"""
    improve torch.repeat / torch.expand function
    eg.1    x.shape = (4, 5, 6, 7); insert_dim = -1; n_times=666
            y = repeat_at(x, insert_dim, n_times)
            y.shape = (4, 5, 6, 7, 666)

    eg.2    x.shape = (4, 5, 6, 7); insert_dim = +1; n_times=666
            y = repeat_at(x, insert_dim, n_times)
            y.shape = (4, 666, 5, 6, 7)
"""
def repeat_at(tensor, insert_dim, n_times):
    if not isinstance(tensor, torch.Tensor):
        return np_repeat_at(tensor, insert_dim, n_times)
    tensor = tensor.unsqueeze(insert_dim)
    shape = list(tensor.shape)
    assert shape[insert_dim] == 1
    shape[insert_dim] = n_times
    return tensor.expand(*shape)

def np_repeat_at(array, insert_dim, n_times):
    array = np.expand_dims(array, insert_dim)
    return array.repeat(axis=insert_dim, repeats=n_times)


def copy_clone(x):
    if x is None:
        return None
    return (
        x.clone()
        if hasattr(x, "clone")
        else x.copy()
        if hasattr(x, "copy")
        else copy.deepcopy(x)
    )


"""
    improve np.reshape and torch.view function
    If a dim is assigned with 0, it will keep its original dimension
    eg.1    x.shape = (4, 5, 6, 7); new_shape = [0, 0, -1]
            y = my_view(x, new_shape)
            y.shape = (4, 5, 6*7)

    eg.2    x.shape = (4, 5, 6, 7); new_shape = [-1, 0, 0]
            y = my_view(x, new_shape)
            y.shape = (4*5, 6, 7)

    eg.3    x.shape = (4, 5, 6); new_shape = [0, 0, -1, 3]
            y = my_view(x, new_shape)
            y.shape = [4, 5, 2, 3]

    eg.4    x.shape = (3, 4, 5, 6); new_shape = [0, 2, -1, 0, 0]
            y = my_view(x, new_shape)
            y.shape = [3, 2, 2, 5, 6]


"""
def my_view(x, shape):
    if -1 in shape[1:-1]: return my_view_test(x, shape)
    reverse_lookup = True if shape[0] == -1 else False
    if not reverse_lookup:
        for i, dim in enumerate(shape):
            if dim == 0:
                shape[i] = x.shape[i]
    else:
        for i in range(len(shape)):
            ni = -(i + 1)  # iter -1,-2,-3,...
            dim = shape[ni]
            if dim == 0:
                shape[ni] = x.shape[ni]
    if isinstance(x, np.ndarray):
        return x.reshape(*shape)
    return x.view(*shape)

def my_view_test(x, shape):
    # fill both way until meet -1 
    for i, dim in enumerate(shape):
        if dim == 0: shape[i] = x.shape[i]
        elif dim == -1: break
    for i in range(len(shape)):
        ni = -(i + 1); dim = shape[ni]
        if dim == 0: shape[ni] = x.shape[ni]
        elif dim == -1: break
    # print(shape)
    if isinstance(x, np.ndarray):
        return x.reshape(*shape)
    return x.view(*shape)


def add_onehot_id_at_last_dim(x):
    _hot_dim = x.shape[-2]
    _identity = np.tile(np.eye(_hot_dim), (*x.shape[:-2], 1, 1))
    return np.concatenate((x, _identity), -1)


"""
    numpy corresponding to torch.nn.functional.one_hot
"""
def np_one_hot(x, n):
    return np.eye(n)[x]


def add_obs_container_subject(container_emb, subject_emb, div):
    # for subject, add one-hot embedding of its group
    n_container = container_emb.shape[1]
    subject_belonging_info = np_one_hot(div, n_container)
    subject_out_emb = np.concatenate((subject_emb, subject_belonging_info), -1)
    # for container, add add multi-hot embedding of its subjects
    container_multihot = np.concatenate(
        [np.expand_dims((div == nth_container).astype(np.long), 1) 
            for nth_container in range(n_container)],
        1,
    )
    container_out_emb = np.concatenate((container_emb, container_multihot), -1)
    return container_out_emb, subject_out_emb



"""
    Function decorate, 
    Turning numpy array to torch.Tensor, then put it on the right GPU / CPU
"""
def Args2tensor(f):
    # if not cfg.init: cfg.read_cfg()
    def _2tensor(x):
        if isinstance(x, torch.Tensor):
            return x.to(cfg.device)
        elif isinstance(x, np.ndarray):
            if (not cfg.use_float64) and x.dtype == np.float64:
                x = x.astype(np.float32)
            if cfg.use_float64 and x.dtype == np.float32:
                x = x.astype(np.float64)
            return torch.from_numpy(x).to(cfg.device)
        elif isinstance(x, dict):
            y = {}
            for key in x:
                y[key] = _2tensor(x[key])
            return y
        else:
            return x

    @wraps(f)
    def decorated(*args, **kwargs):
        for key in kwargs:
            kwargs[key] = _2tensor(kwargs[key])
        return f(*(_2tensor(arg) for arg in args), **kwargs)

    return decorated




def Return2numpy(f):

    def _2cpu2numpy(x):
        return (
            None
            if x is None
            else x
            if not isinstance(x, torch.Tensor)
            else x.detach().cpu().numpy()
            if x.requires_grad
            else x.cpu().numpy()
        )

    @wraps(f)
    def decorated(*args, **kwargs):
        ret_tuple = f(*args, **kwargs)
        if isinstance(ret_tuple, tuple):
            return (_2cpu2numpy(ret) for ret in ret_tuple)
        else:
            return _2cpu2numpy(ret_tuple)

    return decorated


"""
    Function decorate, 
    Turning numpy array to torch.Tensor, then put it on the right GPU / CPU,
    When returning, convert all torch.Tensor to numpy array
"""
def Args2tensor_Return2numpy(f):
    def _2tensor(x):
        if isinstance(x, torch.Tensor):
            return x.to(cfg.device)
        elif isinstance(x, np.ndarray):
            if (not cfg.use_float64) and x.dtype == np.float64:
                x = x.astype(np.float32)
            if cfg.use_float64 and x.dtype == np.float32:
                x = x.astype(np.float64)
            return torch.from_numpy(x).to(cfg.device)
        elif isinstance(x, dict):
            y = {}
            for key in x:
                y[key] = _2tensor(x[key])
            return y
        else:
            return x

    def _2cpu2numpy(x):
        return (
            None
            if x is None
            else x
            if not isinstance(x, torch.Tensor)
            else x.detach().cpu().numpy()
            if x.requires_grad
            else x.cpu().numpy()
        )

    @wraps(f)
    def decorated(*args, **kwargs):
        for key in kwargs:
            kwargs[key] = _2tensor(kwargs[key])
        ret_tuple = f(*(_2tensor(arg) for arg in args), **kwargs)
        return (_2cpu2numpy(ret) for ret in ret_tuple)

    return decorated


"""
    Turning numpy array to torch.Tensor, then put it on the right GPU / CPU,
"""


def _2cpu2numpy(x):
    return (
        None
        if x is None
        else x
        if not isinstance(x, torch.Tensor)
        else x.detach().cpu().numpy()
        if x.requires_grad
        else x.cpu().numpy()
    )


"""
    convert torch.Tensor to numpy array
"""


def _2tensor(x):
    # if not cfg.init: cfg.read_cfg()
    if isinstance(x, torch.Tensor):
        return x.to(cfg.device)
    elif isinstance(x, np.ndarray):
        if (not cfg.use_float64) and x.dtype == np.float64:
            x = x.astype(np.float32)
        if cfg.use_float64 and x.dtype == np.float32:
            x = x.astype(np.float64)
        return torch.from_numpy(x).to(cfg.device)
    elif isinstance(x, dict):
        y = {}
        for key in x:
            y[key] = _2tensor(x[key])
        return y
    else:
        return x


def pad_vec_array(arr_list, max_len):
    # init to NaNs
    res = np.zeros(shape=(len(arr_list), max_len), dtype=np.double) + np.nan
    for i in range(len(arr_list)):
        if arr_list[i] is None:
            continue
        res[i, : len(arr_list[i])] = arr_list[i]
    return res


def one_hot_with_nan_np(tensr, num_classes):
    tensr = tensr.copy()
    tensr[np.isnan(tensr)] = num_classes
    Res_1MoreCol = np_one_hot(tensr.astype(np.long), num_classes + 1)
    return Res_1MoreCol[..., :-1]


def one_hot_with_nan(tensr, num_classes):
    if isinstance(tensr, np.ndarray):
        return one_hot_with_nan_np(tensr, num_classes)
    tensr = tensr.clone()
    tensr[torch.isnan(tensr)] = num_classes
    Res_1MoreCol = F.one_hot(tensr.long(), num_classes + 1)
    return Res_1MoreCol[..., :-1]


def scatter_with_nan(tensr, num_classes, out_type="binary"):
    res = one_hot_with_nan(tensr, num_classes)
    res = res.sum(-2)
    if out_type == "bool":
        res = res != 0
    return res


def process_space(space):
    # starcraft 环境无须特殊处理
    if not ("Box" in space["obs_space"] or "Discrete" in space["act_space"]):
        return space

    # 其他环境需要进行格式转换
    import re

    obs_dim = int(
        re.findall(
            re.compile(r"Box[(]-inf, inf, [(](.*?)[,)]", re.S), space["obs_space"]
        )[0]
    )
    print(space["obs_space"])
    space_ = {}
    space_["obs_space"] = {}
    space_["act_space"] = {}
    space_["obs_space"]["state_shape"] = 8
    space_["obs_space"]["obs_shape"] = obs_dim
    space_["act_space"]["n_actions"] = 8
    space_["obs_space"] = str(space_["obs_space"])
    space_["act_space"] = str(space_["act_space"])
    return space_


class Policy_shift_observer(object):
    def __init__(self, act_range, act_num):
        self.act_range = act_range  # 15
        self.act_num = act_num  # 3
        self.act_cnt_array = np.zeros(shape=(act_num, act_range))
        self.rate = None
        self.rate_history = None

    def new_sample(self, act):
        act_rec = act.shape[0]
        for act_index in range(self.act_num):
            for act_nth in range(self.act_range):
                self.act_cnt_array[act_index, act_nth] = torch.sum(
                    (act[:, act_index] == act_nth).long()
                )
        self.rate = self.act_cnt_array / act_rec
        if self.rate_history is None:
            self.rate_history = self.rate
        else:
            self.rate_history = self.rate_history * 0.9 + self.rate * 0.1
        print("rate", self.rate)
        # conclusion: the action distribution is not reinforced because the rewards signal is too weak.


"""
    Get the hash code string of an array, 
    compatable for numpy array and torch.tensor
"""
def __hash__(x):
    import hashlib
    md5 = hashlib.md5()  # ignore
    # if isinstance(x, str):
    #     md5.update(x)
    #     return md5.hexdigest()
    if hasattr(x, "cpu"):
        md5.update(x.detach().cpu().numpy().data.tobytes())
        return md5.hexdigest()
    elif hasattr(x, "numpy"):
        md5.update(x.numpy().data.tobytes())
        return md5.hexdigest()
    elif hasattr(x, "data"):
        md5.update(x.data.tobytes())
        return md5.hexdigest()
    else:
        try:
            md5.update(x.encode("utf-8"))
            return md5.hexdigest()
        except:
            return str(x)


def __hashm__(*args):
    import hashlib

    md5 = hashlib.md5()  # ignore
    for arg in args:
        x = arg
        if hasattr(x, "cpu"):
            md5.update(x.detach().cpu().numpy().data.tobytes())
        elif hasattr(x, "numpy"):
            md5.update(x.numpy().data.tobytes())
        elif hasattr(x, "data"):
            md5.update(x.data.tobytes())
        else:
            try:
                md5.update(x.encode("utf-8"))
            except:
                md5.update(str(x).encode("utf-8"))
    return md5.hexdigest()


"""
    Get the hash code string of the pytorch network parameters
    eg. 
        __hashn__(mlp_module.parameters())
"""
def __hashn__(generator):
    import hashlib

    md5 = hashlib.md5()  # ignore
    for arg in generator:
        x = arg.data
        if hasattr(x, "cpu"):
            md5.update(x.detach().cpu().numpy().data.tobytes())
        elif hasattr(x, "numpy"):
            md5.update(x.numpy().data.tobytes())
        elif hasattr(x, "data"):
            md5.update(x.data.tobytes())
        else:
            try:
                md5.update(x.encode("utf-8"))
            except:
                md5.update(str(x).encode("utf-8"))
    return md5.hexdigest()


"""
    numpy version of softmax
"""


def np_softmax(x, axis=None):
    # compute in log space for numerical stability
    return np.exp(x - logsumexp(x, axis=axis, keepdims=True))


"""
    numpy version of logsumexp
"""


def logsumexp(a, axis=None, keepdims=False, return_sign=False):
    a_max = np.amax(a, axis=axis, keepdims=True)
    if a_max.ndim > 0:
        a_max[~np.isfinite(a_max)] = 0
    elif not np.isfinite(a_max):
        a_max = 0
    tmp = np.exp(a - a_max)
    # suppress warnings about log of zero
    with np.errstate(divide="ignore"):
        s = np.sum(tmp, axis=axis, keepdims=keepdims)
        if return_sign:
            sgn = np.sign(s)
            s *= sgn  # /= makes more sense but we need zero -> zero
        out = np.log(s)
    if not keepdims:
        a_max = np.squeeze(a_max, axis=axis)
    out += a_max
    if return_sign:
        return out, sgn
    else:
        return out


"""
函数说明：在有限的、不均衡的多标签数据集中，按照预设的比例，取出尽可能多的样本
"""


def sample_balance(x, y, n_class, weight=None):
    if weight is None:
        weight = torch.ones(n_class, device=x.device)
    else:
        weight = torch.Tensor(weight).to(x.device)
    n_instance = torch.zeros(n_class, device=x.device)
    indices = [None] * n_class
    for i in range(n_class):
        indices[i] = torch.where(y == i)[0]
        n_instance[i] = len(indices[i])
    ratio = n_instance / weight
    bottle_neck = torch.argmin(n_instance / weight)
    r = ratio[bottle_neck]
    n_sample = (r * weight).long()
    # print(n_instance, n_sample)
    new_indices = [indices[i][torch.randperm(n_sample[i])] for i in range(n_class)]
    # print(new_indices)
    new_indices_ = torch.cat(new_indices)
    assert len(new_indices_) == sum(n_sample)
    return x[new_indices_], y[new_indices_]


"""
    gather tensor with index, 
    regarding all right hand dimensions as dimensions need to be gathered
    eg.1
        src = torch.Tensor([[[ 0,  1,  2], [ 3,  4,  5]],
                            [[ 6,  7,  8], [ 9, 10, 11]],
                            [[12, 13, 14], [15, 16, 17]]])
        index = torch.Tensor([[0], [1], [0]])
        src.shape = (3, 2, 3)
        src.shape = (3, 1)
        res = gather_righthand(src,index)
        res.shape = (3, 1, 3)
        res= tensor([[[ 0.,  1.,  2.]],
                     [[ 9., 10., 11.]],
                     [[12., 13., 14.]]])
    eg.2
        src.shape   = (64, 16, 8, 88, 888)
        index.shape = (64, 5)
        res = gather_righthand(src,index)
        res.shape   = (64, 5,  8, 88, 888)

    eg.3
        src.shape   = (64, 16, 88, 888)
        index.shape = (64, 777)
        res = gather_righthand(src,index)
        res.shape   = (64, 777,  88, 888)

"""
def gather_righthand(src, index, check=True):
    if not isinstance(src, torch.Tensor):
        return np_gather_righthand(src, index, check)
    index = index.long()
    i_dim = index.dim()
    s_dim = src.dim()
    t_dim = i_dim - 1
    if check:
        assert s_dim >= i_dim
        assert index.max() <= src.shape[t_dim] - 1
        if index.max() != src.shape[t_dim] - 1:
            print(
                "[gather_righthand] warning, index max value does not match src target dim"
            )
        assert (
            src.shape[t_dim] != index.shape[t_dim]
        ), "you really want to select %d item out of %d?" % (
            index.shape[t_dim],
            src.shape[t_dim],
        )
        for d in range(0, t_dim):
            assert src.shape[d] == index.shape[d]
    index_new_shape = list(src.shape)
    index_new_shape[t_dim] = index.shape[t_dim]
    for _ in range(i_dim, s_dim):
        index = index.unsqueeze(-1)
    index_expand = index.expand(index_new_shape)  # only this two line matters
    return torch.gather(
        src, dim=t_dim, index=index_expand
    )  # only this two line matters


def np_gather_righthand(src, index, check=True):
    index = index.astype(np.long)
    dim = lambda x: len(x.shape)
    i_dim = dim(index)
    s_dim = dim(src)
    t_dim = i_dim - 1
    if check:
        assert s_dim >= i_dim
        assert index.max() <= src.shape[t_dim] - 1
        if index.max() != src.shape[t_dim] - 1:
            print(
                "[gather_righthand] warning, index max value does not match src target dim"
            )
        assert (
            src.shape[t_dim] != index.shape[t_dim]
        ), "you really want to select %d item out of %d?" % (
            index.shape[t_dim],
            src.shape[t_dim],
        )
        for d in range(0, t_dim):
            assert src.shape[d] == index.shape[d]
    tile_shape = np.array(src.shape)  # warning: careful when moving to pytorch
    tile_shape[: (t_dim + 1)] = 1
    for _ in range(i_dim, s_dim):
        index = np.expand_dims(index, -1)
    index_expand = np.tile(
        index, tile_shape
    )  # index.expand(index_new_shape)            # only this two line matters
    return np.take_along_axis(arr=src, indices=index_expand, axis=t_dim)
    # return torch.gather(src, dim=t_dim, index=index_expand) # only this two line matters


def distance_matrix(A):
    assert A.shape[-1] == 2  # assert 2D situation
    n_subject = A.shape[-2]  # is 2
    A = np.repeat(np.expand_dims(A, -2), n_subject, axis=-2)  # =>(64, 100, 100, 2)
    At = np.swapaxes(A, -2, -3)  # =>(64, 100, 100, 2)
    dis = At - A  # =>(64, 100, 100, 2)
    dis = np.linalg.norm(dis, axis=-1)
    return dis

def delta_matrix(A):
    n_subject = A.shape[-2]  # is 2
    A = np.repeat(np.expand_dims(A, -2), n_subject, axis=-2)  # =>(64, 100, 100, 2)
    At = np.swapaxes(A, -2, -3)  # =>(64, 100, 100, 2)
    delta = At - A  # =>(64, 100, 100, 2)
    return delta

def np_normalize_last_dim(mat):
    return mat / np.expand_dims(np.linalg.norm(mat, axis=-1) + 1e-16, axis=-1)

def dir2rad(delta_pos):
    result = np.empty(delta_pos.shape[:-1], dtype=complex)
    result.real = delta_pos[..., 0]
    result.imag = delta_pos[..., 1]
    rad_angle = np.angle(result)
    return rad_angle


def dir3d_rad(delta_pos):
    assert delta_pos.shape[-1]==3
    xy = delta_pos[..., :2]
    r1 = dir2rad(xy)
    xy_norm = np.linalg.norm(xy, axis=-1)
    r2 = dir2rad(np.stack((xy_norm, delta_pos[..., 2]),-1))
    return np.stack((r1,r2), axis=-1)
    

def reg_deg(deg):
    return (deg + 180) % 360 - 180

# make angles comparable
def reg_deg_at(rad, ref):
    return reg_deg(rad-ref) + ref


def reg_rad(rad):
    # it's OK to show "RuntimeWarning: invalid value encountered in remainder"
    return (rad + np.pi) % (2 * np.pi) - np.pi

# make angles comparable
def reg_rad_at(rad, ref):
    return reg_rad(rad-ref) + ref



def zeros_like_except_dim(array, except_dim, n):
    shape_ = list(array.shape)
    shape_[except_dim] = n
    return torch.zeros(size=shape_, device=array.device, dtype=array.dtype)


def pad_at_dim(array, dim, n):
    extra_n = n-array.shape[dim]
    padding = zeros_like_except_dim(array, except_dim=dim, n=extra_n)
    return torch.cat((array, padding), axis=dim)

def stack_vec_with_padding(arr_list):
    _len = [arr.len() for arr in arr_list]
    max_len = max(_len)
    n_subject = arr_list.len()
    dtype = arr_list[0].dtype
    arr_np = np.zeros(shape=(n_subject, max_len), dtype=dtype)
    for i, arr in enumerate(arr_list):
        arr_np[i,:_len[i]] = arr
    return arr_np


def objdump(obj):
    import pickle
    with open('objdump.tmp', 'wb+') as f:
        pickle.dump(obj, f)
    return

def objload():
    import pickle, os
    if not os.path.exists('objdump.tmp'): 
        return
    with open('objdump.tmp', 'rb') as f:
        return pickle.load(f)


import flock, os, torch, uuid, time
from atexit import register

class GpuShareUnit():
    def __init__(self, which_gpu, lock_path=None, manual_gpu_ctl=True, gpu_party=''):
        self.device = which_gpu
        self.manual_gpu_ctl = True
        self.lock_path=lock_path
        self.gpu_party = gpu_party
        self.experiment_uuid = uuid.uuid1().hex + '\n'
        self.n_gpu_process_online = 1
        if gpu_party == 'off':
            self.manual_gpu_ctl = False
        if self.lock_path is None: 
            self.lock_path = os.path.expanduser('~/GpuLock')
        if not os.path.exists(self.lock_path): os.makedirs(self.lock_path)
        register(self.unregister_uuid_)
        
    def __del__(self):
        self.unregister_uuid_()
        try: self.gpu_lock.__exit__(None,None,None)
        except:pass
        try: self.gpu_lock_file.close()
        except:pass

    def __enter__(self):
        self.get_gpu_lock()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.release_gpu_lock()

    def get_gpu_lock(self):
        if self.manual_gpu_ctl:
            print('Waiting for GPU %s %s...'%(self.device, self.gpu_party), end='', flush=True)
            self.gpu_lock, self.gpu_lock_file = (None, None)
            self.n_gpu_process_online = self.register_uuid_()
            self.gpu_lock_file = open(self.lock_path+'/lock_gpu_%s_%s.glock'%(self.device, self.gpu_party), 'w+')
            self.gpu_lock = flock.Flock(self.gpu_lock_file, flock.LOCK_EX)
            self.gpu_lock.__enter__()
            print('Get GPU, currently shared with %d process!'%self.n_gpu_process_online)
        return

    def release_gpu_lock(self):
        if self.manual_gpu_ctl:
            if self.n_gpu_process_online >1: 
                torch.cuda.empty_cache()
                self.gpu_lock.__exit__(None,None,None)
                self.gpu_lock_file.close()
            else:
                print('不共享GPU')
        return
    
    def register_uuid_(self):
        try:
            flag = 'w+' if not os.path.exists(self.lock_path+'/lock_gpu_%s_%s.register'%(self.device, self.gpu_party)) else 'r+'
            with open(self.lock_path+'/lock_gpu_%s_%s.register'%(self.device, self.gpu_party), mode=flag) as gpu_register_file:
                _lock = flock.Flock(gpu_register_file, flock.LOCK_EX); _lock.__enter__()
                lines = gpu_register_file.readlines()
                if not any([line==self.experiment_uuid for line in lines]):
                    lines.append(self.experiment_uuid)
                    gpu_register_file.seek(0); gpu_register_file.truncate(0)
                    gpu_register_file.writelines(lines)
                    gpu_register_file.flush()
                _lock.__exit__(None,None,None)
                return len(lines)
        except:
            print('GPU 队列异常!')
            return 999

    def unregister_uuid_(self):
        flag = 'w+' if not os.path.exists(self.lock_path+'/lock_gpu_%s_%s.register'%(self.device, self.gpu_party)) else 'r+'
        with open(self.lock_path+'/lock_gpu_%s_%s.register'%(self.device, self.gpu_party), mode=flag) as gpu_register_file:
            _lock = flock.Flock(gpu_register_file, flock.LOCK_EX); _lock.__enter__()
            lines = gpu_register_file.readlines()
            gpu_register_file.seek(0); gpu_register_file.truncate(0)
            gpu_register_file.writelines([line for line in lines if line!=self.experiment_uuid])
            gpu_register_file.flush()
            _lock.__exit__(None,None,None)
            print('unregister')




import paramiko, os, time
from UTILS.colorful import print亮紫, print亮靛
class ChainVar(object):
    def __init__(self, chain_func, chained_with):
        self.chain_func = chain_func
        self.chained_with = chained_with

class DataCentralServer(object): # ADD_TO_CONF_SYSTEM //DO NOT remove this comment//
    addr = 'None'
    usr = 'None'
    pwd = 'None'


# great thank to skoll for sharing this at stackoverflow:
# https://stackoverflow.com/questions/4409502/directory-transfers-with-paramiko
class MySFTPClient(paramiko.SFTPClient):
    def put_dir(self, source, target, ignore_list=[]):
        ''' Uploads the contents of the source directory to the target path. The
            target directory needs to exists. All subdirectories in source are 
            created under target.
        '''
        for item in os.listdir(source):
            if item in ignore_list: continue
            if os.path.isfile(os.path.join(source, item)):
                # print亮靛('uploading: %s --> %s'%(os.path.join(source, item),'%s/%s' % (target, item)))
                self.put(os.path.join(source, item), '%s/%s' % (target, item))
            else:
                self.mkdir('%s/%s' % (target, item), ignore_existing=True)
                self.put_dir(os.path.join(source, item), '%s/%s' % (target, item), ignore_list)

    def mkdir(self, path, mode=511, ignore_existing=False):
        ''' Augments mkdir by adding an option to not fail if the folder exists  '''
        try:
            super(MySFTPClient, self).mkdir(path, mode)
        except IOError as e:
            if e.__class__ == FileNotFoundError:
                raise

            if ignore_existing:
                pass
            else:
                raise


def get_ssh_sftp(addr, usr, pwd):
    ssh = paramiko.SSHClient() 
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.load_host_keys(os.path.expanduser(os.path.join("~", ".ssh", "known_hosts")))
    port = 22
    if ':' in addr: addr, port = addr.split(':')
    ssh.connect(addr, username=usr, password=pwd, port=port)
    sftp = MySFTPClient.from_transport(ssh.get_transport())
    return ssh, sftp


def upload_experiment_results(cfg): # shell it to catch error
    try: upload_experiment_results_(cfg)
    except: pass

def upload_experiment_results_(cfg):
    path = cfg.logdir
    name = cfg.note
    try:
        addr = DataCentralServer.addr     # ssh ubuntu address
        usr = DataCentralServer.usr      # ubuntu user
        pwd = DataCentralServer.pwd      # ubuntu password
        assert addr != 'None' and (addr is not None)
        assert usr != 'None' and (usr is not None)
        assert pwd != 'None' and (pwd is not None)
    except:
        print('No experiment data central server is configured, 没有配置中央日志服务器')
        return
    remote_path = '/home/%s/CenterHmp/'%usr
    ssh = paramiko.SSHClient() 
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.load_host_keys(os.path.expanduser(os.path.join("~", ".ssh", "known_hosts")))
    ssh.connect(addr, username=usr, password=pwd)
    put_str = '[%s] [%s] %s'%(cfg.note, time.strftime("%Y-%m-%d-%H:%M:%S", time.localtime()), str(cfg.machine_info).replace('\'',''))
    ssh.exec_command(command='echo -e "%s" >> %s/active.log'%(put_str, remote_path), timeout=1)
    sftp = MySFTPClient.from_transport(ssh.get_transport())
    print亮紫('uploading results: %s --> %s'%(path, '%s/%s'%(remote_path, name)))
    sftp.mkdir(remote_path, ignore_existing=True)
    sftp.mkdir('%s/%s'%(remote_path, name), ignore_existing=True)
    sftp.put_dir(path, '%s/%s'%(remote_path, name))
    sftp.close()
    print亮紫('upload complete')
    


import os
import numpy as np
import matplotlib
from config import GlobalConfig
# 设置matplotlib正常显示中文和负号
# matplotlib.rcParams['font.sans-serif']=['SimHei']   # 用黑体显示中文
# matplotlib.rcParams['axes.unicode_minus']=False     # 正常显示负号
# plt.ion()
StandardPlotFig = 1
ComparePlotFig = 2
# from pylab import *
class rec_family(object):
    def __init__(self, colorC=None, draw_mode='Native'):
        self.name_list = []
        self.line_list = []
        self.line_plot_handle = []
        self.line_plot_handle2 = []
        self.subplots = {}
        self.subplots2 = {}
        self.working_figure_handle = None
        self.working_figure_handle2 = None
        self.smooth_line = False
        self.colorC = 'k' if colorC is None else colorC
        self.Working_path = 'Testing-beta'
        self.image_num = -1
        self.draw_mode = draw_mode
        self.vis_95percent = True
        logdir = GlobalConfig.logdir
        self.plt = None
        if not os.path.exists(logdir):
            os.makedirs(logdir)
        if self.draw_mode == 'Web':
            import matplotlib.pyplot as plt, mpld3
            self.html_to_write = '%s/html.html'%logdir
            self.plt = plt; self.mpld3 = mpld3
        elif self.draw_mode =='Native':
            import matplotlib.pyplot as plt
            plt.ion()
            self.plt = plt
        elif self.draw_mode =='Img':
            matplotlib.use('Agg')
            import matplotlib.pyplot as plt
            self.plt = plt
            self.img_to_write = '%s/rec.jpg'%logdir
        else:
            assert False

    def rec_init(self, colorC=None):
        if colorC is not None: self.colorC = colorC
        return

    def rec(self, var, name):
        if name in self.name_list:
            pass
        else:
            self.name_list.append(name)
            self.line_list.append([])  #新建一个列表
            self.line_plot_handle.append(None)
            self.line_plot_handle2.append(None)
        
        index = self.name_list.index(name)
        self.line_list[index].append(var)
    
    # This function is ugly because it is translated from MATLAB
    def rec_show(self):
        image_num = len(self.line_list)  #一共有多少条曲线
        #画重叠曲线，如果有的话
        
        if self.working_figure_handle is None:
            self.working_figure_handle = self.plt.figure(StandardPlotFig, figsize=(12, 6), dpi=100)
            if self.draw_mode == 'Native': 
                self.working_figure_handle.canvas.set_window_title(self.Working_path)
                self.plt.show()

        
        rows = 1
        #检查是否有时间轴，若有，做出修改
        flag_time_e = 0
        encountered = 0 # 有时间轴
        if 'time' in self.name_list:
            time_index = self.name_list.index('time')
            image_num = image_num - 1
            flag_time_e = 1
            encountered = 0 
        
        if image_num >= 3:
            rows = 2#大与3张图，则放两行

        if flag_time_e>0:
            image_num = image_num + 1
        
        cols = int(np.ceil(image_num/rows))#根据行数求列数
        if self.image_num!=image_num:
            # 需要刷新布局，所有已经绘制的图作废
            self.subplots = {}
            self.working_figure_handle.clf()
            for q,handle in enumerate(self.line_plot_handle): 
                self.line_plot_handle[q] = None

        self.image_num = image_num
        # # 太多了的话直接在二图展示
        if image_num < 12:
            for index in range(image_num):
                if flag_time_e>0:
                    if time_index == index:
                        encountered = 1 # 有时间轴
                        continue
                # 有时间轴时，因为不绘制时间，所以少算一个subplot:
                subplot_index = index if encountered > 0 else index+1
                subplot_name = '%d,%d,%d'%(rows,cols,subplot_index)
                if subplot_name in self.subplots: 
                    target_subplot = self.subplots[subplot_name]
                else:
                    target_subplot = self.working_figure_handle.add_subplot(rows,cols,subplot_index)
                    self.subplots[subplot_name] = target_subplot

                _xdata_ = np.arange(len(self.line_list[index]), dtype=np.double)
                _ydata_ = np.array(self.line_list[index], dtype=np.double)
                if flag_time_e>0:
                    _xdata_ = np.array(self.line_list[time_index], dtype=np.double)
                # plt.plot(x,y,ls,lw,c,marker,markersize,markeredgecolor,markerfacecolor,label)
                # **x：**横坐标；**y：**纵坐标；**ls或linestyle：**线的形式（‘-’，‘–’，‘：’和‘-.’）；**lw（或linewidth）：**线的宽度；**c：**线的颜色；**marker：**线上点的形状；**markersize或者ms：**标记的尺寸，浮点型；**markerfacecolor：**点的填充色；**markeredgecolor：标记的边沿颜色label：**文本标签
                if (self.line_plot_handle[index] is None):# || ~isvalid(self.line_plot_handle[index])):
                        if flag_time_e>0:
                            self.line_plot_handle[index]  =  target_subplot.plot(self.line_list[time_index],self.line_list[index],lw=1,c=self.colorC)
                        else:
                            self.line_plot_handle[index], =  target_subplot.plot(self.line_list[index], lw=1, c=self.colorC)
                        
                else:
                    if flag_time_e>0:
                        self.line_plot_handle[index].set_data((self.line_list[time_index],self.line_list[index]))
                    else:
                        xdata = np.arange(len(self.line_list[index]), dtype=np.double)
                        ydata = np.array(self.line_list[index], dtype=np.double)
                        self.line_plot_handle[index].set_data((xdata,ydata))


                if 'of=' in self.name_list[index]:
                    #把等号替换成空格
                    name_tmp = self.name_list[index]
                    name_tmp = name_tmp.replace('=',' ')
                    target_subplot.set_title(name_tmp)
                    target_subplot.set_xlabel('time')
                    target_subplot.set_ylabel(name_tmp)
                    target_subplot.grid(visible=True)
                else:
                    target_subplot.set_title(self.name_list[index])
                    target_subplot.set_xlabel('time')
                    target_subplot.set_ylabel(self.name_list[index])
                    target_subplot.grid(visible=True)

                #画图时边界留出一点空间
                # axis tight
                # target_subplot.relim()
                    
                limx1 = _xdata_.min() #target_subplot.dataLim.xmin
                limx2 = _xdata_.max() #target_subplot.dataLim.xmax
                limy1 = _ydata_.min() #min(self.line_list[index])
                limy2 = _ydata_.max() #max(self.line_list[index])

                if len(_ydata_)>220 and self.vis_95percent:
                    limy1 = np.percentile(_ydata_, 3, interpolation='midpoint') # 5%
                    limy2 = np.percentile(_ydata_, 97, interpolation='midpoint') # 95%

                if limx1 != limx2 and limy1!=limy2:
                    # limx1,limy1,limx2,limy2 = target_subplot.dataLim
                    meany = limy1/2 + limy2/2
                    limy1 = (limy1 - meany)*1.2+meany
                    limy2 = (limy2 - meany)*1.2+meany
                    target_subplot.set_ylim(limy1,limy2)

                    meanx = limx1/2 + limx2/2
                    limx1 = (limx1 - meanx)*1.1+meanx
                    limx2 = (limx2 - meanx)*1.1+meanx
                    target_subplot.set_xlim(limx1,limx2)
                elif limx1 != limx2:
                    meanx = limx1/2 + limx2/2
                    limx1 = (limx1 - meanx)*1.1+meanx
                    limx2 = (limx2 - meanx)*1.1+meanx
                    target_subplot.set_xlim(limx1,limx2)
            
                # plt.draw()
        # ##################################################
        # ##################################################
        

        # ##
        # #画重叠曲线，如果有的话
        draw_advance_fig = False
        for name in self.name_list:
            if 'of=' in name: draw_advance_fig = True

        if draw_advance_fig:

            #画重叠曲线，如果有的话
            if self.working_figure_handle2 is None:
                self.working_figure_handle2 = self.plt.figure(ComparePlotFig, figsize=(12, 6), dpi=100)
                if self.draw_mode == 'Native': 
                    self.working_figure_handle2.canvas.set_window_title('Working')
                    self.plt.show()
            
            group_name = []
            group_member = []
            
            for index in range(image_num):
                if 'of=' not in self.name_list[index]:
                    continue    #没有的直接跳过
                # 找出组别
                res = self.name_list[index].split('of=')
                g_name_ = res[0]
                
                if g_name_ in group_name: # any(strcmp(group_name, g_name_)):
                    i = group_name.index(g_name_)
                    group_member[i].append(index)# ,index]  ##ok<*AGROW>
                else:
                    group_name.append(g_name_)
                    group_member.append([index])
                
                
            
            num_group = len(group_name)
            image_num_multi = num_group
            if image_num_multi >= 3:
                rows = 2#大与3张图，则放两行
            else:
                rows = 1
            
            cols = int(np.ceil(image_num_multi/rows))#根据行数求列数
            
            for i in range(num_group):

                subplot_index = i+1
                subplot_name = '%d,%d,%d'%(rows,cols,subplot_index)
                if subplot_name in self.subplots2: 
                    target_subplot = self.subplots2[subplot_name]
                else:
                    target_subplot = self.working_figure_handle2.add_subplot(rows,cols,subplot_index)
                    self.subplots2[subplot_name] = target_subplot

                tar_true_name=group_name[i]
                num_member = len(group_member[i])
                
                for j in range(num_member):
                    index = group_member[i][j]
                    name_tmp = self.name_list[index]
                    name_tmp = name_tmp.replace('=',' ')
                    if self.smooth_line:
                        target = smooth(self.line_list[index],20) 
                    else:
                        target = self.line_list[index]
                    if (self.line_plot_handle2[index] is None):
                        if flag_time_e>0:
                            self.line_plot_handle2[index]  =  target_subplot.plot(self.line_list[time_index],self.line_list[index],lw=1,label=name_tmp)
                        else:
                            self.line_plot_handle2[index], =  target_subplot.plot(self.line_list[index], lw=1, label=name_tmp)

                    else:
                        if flag_time_e>0:
                            self.line_plot_handle2[index].set_data((self.line_list[time_index],self.line_list[index]))
                        else:
                            xdata = np.arange(len(self.line_list[index]), dtype=np.double)
                            ydata = np.array(self.line_list[index], dtype=np.double)
                            self.line_plot_handle2[index].set_data((xdata,ydata))

                #标题
                target_subplot.set_title(tar_true_name)
                target_subplot.set_xlabel('time')
                target_subplot.set_ylabel(tar_true_name)
                target_subplot.relim()

                limx1 = target_subplot.dataLim.xmin
                limx2 = target_subplot.dataLim.xmax
                limy1 = target_subplot.dataLim.ymin
                limy2 = target_subplot.dataLim.ymax
                # limx1,limy1,limx2,limy2 = target_subplot.dataLim
                if limx1 != limx2 and limy1!=limy2:
                    meany = limy1/2 + limy2/2
                    limy1 = (limy1 - meany)*1.2+meany
                    limy2 = (limy2 - meany)*1.2+meany
                    target_subplot.set_ylim(limy1,limy2)
                    meanx = limx1/2 + limx2/2
                    limx1 = (limx1 - meanx)*1.05+meanx
                    limx2 = (limx2 - meanx)*1.05+meanx
                    target_subplot.set_xlim(limx1,limx2)
                    target_subplot.grid(visible=True)
                    target_subplot.legend(loc='best')
                elif limx1 != limx2:
                    meanx = limx1/2 + limx2/2
                    limx1 = (limx1 - meanx)*1.1+meanx
                    limx2 = (limx2 - meanx)*1.1+meanx
                    target_subplot.set_xlim(limx1,limx2)
        # now end
        self.plt.tight_layout()
        if self.draw_mode == 'Web':
            content = self.mpld3.fig_to_html(self.working_figure_handle)
            with open(self.html_to_write, 'w+') as f:
                f.write(content)
            return
        elif self.draw_mode == 'Native':
            self.plt.pause(0.01)
            return
        elif self.draw_mode == 'Img':
            if self.working_figure_handle is not None: 
                self.working_figure_handle.savefig(self.img_to_write)
                # print('保存图像')



import time, requests, threading, os, atexit
from UTILS.colorful import *


def kill_process(p):
    try:
        print('正在发送terminate命令到进程:', p.pid)
        p.terminate()
        print('正在发送kill命令到进程:', p.pid)
        time.sleep(3)
        p.kill()
    except: pass

def kill_process_and_its_children(p):
    print('开始杀死以下进程及其子进程:',p.pid)
    for child in p.children():
        print('->', child.pid)

    for child in p.children():
        if hasattr(child,'children') and len(child.children())>0:
            kill_process_and_its_children(child)
        else:
            kill_process(child)
    kill_process(p)



def clean_child_process(pid):
    import psutil, time
    parent = psutil.Process(pid)
    kill_process_and_its_children(parent)


def hmp_clean_up():
    from UTILS.exp_upload import upload_experiment_results
    from config import GlobalConfig as cfg
    print亮黄('[main.py] upload results to storage server via SSH')
    if cfg.allow_res_upload: upload_experiment_results(cfg)
    print亮黄('[main.py] kill all children process, then self-terminate.')
    clean_child_process(os.getpid())

def start_periodic_daemon(cfg):
    periodic_thread = threading.Thread(target=periodic_daemon,args=(cfg,))
    periodic_thread.setDaemon(True)
    periodic_thread.start()
    atexit.register(hmp_clean_up)

def periodic_daemon(cfg):
    while True:
        try: 
            print('start periodic_daemon_(cfg)')
            periodic_daemon_(cfg)
            print('end periodic_daemon_(cfg)')
        except AssertionError: 
            hmp_clean_up()
        except BaseException: 
            print('hmp server failed')
            break
        time.sleep(15*60)

def periodic_daemon_(cfg):
    report = {
        'type': 'hmp-client',
        'note': cfg.note,
        'time': time.strftime("%Y-%m-%d-%H:%M:%S", time.localtime()),
        'client_status': 'Running',
        'StartingTime': cfg.machine_info['StartDateTime'],
        'HostIP': cfg.machine_info['HostIP'],
        'ExpUUID': cfg.machine_info['ExpUUID'],
        'RunPath':cfg.machine_info['RunPath'],
        'DockerContainerHash':cfg.machine_info['DockerContainerHash']
    }
    res = requests.post('http://linux.ipv4.fuqingxu.top:11511/',data = report)

    if res.text=='Stop_Now': 
        report['client_status'] = 'Terminate'
        requests.post('http://linux.ipv4.fuqingxu.top:11511/',data = report)
        raise AssertionError('HMP-Center Has Given Terminate Signal!')


# import os
# print(os.getcwd())
import os, sys
import argparse
from VISUALIZE.mcom import *



class RecallProcessThreejs(Process):
    def __init__(self, file_path):
        super(RecallProcessThreejs, self).__init__()
        self.buffer_list = []
        self.file_path = file_path

    def init_threejs(self):
        import threading
        t = threading.Thread(target=self.run_flask, args=(5051,))
        # t = threading.Thread(target=self.run_flask, args=(51241,))
        t.start()

    def run(self):
        self.init_threejs()
        try:
            with open(self.file_path, 'r') as f:
                new_buff_list = f.readlines()
            self.run_handler(new_buff_list)
            while True: time.sleep(1000)
        except KeyboardInterrupt:
            self.__del__()

        self.__del__()

    def run_handler(self, new_buff_list):
        self.buffer_list.extend(new_buff_list)
        # too many, delete with fifo
        if len(self.buffer_list) > 1e9: # 当存储的指令超过十亿后，开始删除旧的
            del self.buffer_list[:len(new_buff_list)]

    def run_flask(self, port):
        from flask import Flask, url_for, jsonify, request, send_from_directory, redirect
        from waitress import serve

        app = Flask(__name__)
        dirname = os.path.dirname(__file__) + '/threejsmod'
        import zlib

        @app.route("/up", methods=["POST"])
        def upvote():
            # dont send too much in one POST, might overload the network traffic
            if len(self.buffer_list)>35000:
                tosend = self.buffer_list[:30000]
                self.buffer_list = self.buffer_list[30000:]
            else:
                tosend = self.buffer_list
                self.buffer_list = []

            # use zlib to compress output command, worked out like magic
            buf = "".join(tosend)
            buf = bytes(buf, encoding='utf8')   
            zlib_compress = zlib.compressobj()
            buf = zlib_compress.compress(buf) + zlib_compress.flush(zlib.Z_FINISH)
            return buf

        @app.route("/<path:path>")
        def static_dirx(path):
            if path=='favicon.ico': 
                return app.send_static_file('%s/files/favicon.ico'%dirname)
            return send_from_directory("%s/"%dirname, path)

        @app.route("/")
        def main_app():
            with open('%s/examples/abc.html'%dirname, 'r', encoding = "utf-8") as f:
                buf = f.read()
            return buf

        print('\n--------------------------------')
        print('JS visualizer online: http://%s:%d'%(get_host_ip(), port))
        print('JS visualizer online (localhost): http://localhost:%d'%(port))
        print('--------------------------------')
        # app.run(host='0.0.0.0', port=port)
        serve(app, threads=8, ipv4=True, ipv6=True, listen='*:%d'%port)



if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='HMP')
    parser.add_argument('-p', '--path', help='directory of chosen file')
    args, unknown = parser.parse_known_args()
    if hasattr(args, 'path'):
        path = args.path
    else:
        assert False, (r"parser.add_argument('-p', '--path', help='The node name is?')")

    load_via_json = (hasattr(args, 'cfg') and args.cfg is not None)
    
    rp = RecallProcessThreejs(path)
    rp.start()
    rp.join()



import socket
def find_free_port():
    from contextlib import closing
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]

def get_host_ip():
    ip = None
    try:
        s=socket.socket(socket.AF_INET,socket.SOCK_DGRAM)
        s.connect(('8.8.8.8',80))
        ip=s.getsockname()[0]
    finally:
        s.close()
    return ip



import argparse, os, time, func_timeout
from shutil import copyfile, copytree, ignore_patterns
from .colorful import *

def secure_chained_vars(default_cfg, new_cfg, vb):
    default_cfg_dict = default_cfg.__dict__
    altered_cv = []
    for key in default_cfg_dict:
        if not key.endswith('_cv'): continue
        o_key = key.replace('_cv','')
        if o_key in new_cfg: continue
        assert hasattr(default_cfg, o_key), ('twin var does not have original')
        # get twin
        chain_var = getattr(default_cfg, key)   
        need_reflesh = False
        for chain_by_var in chain_var.chained_with:
            if chain_by_var in new_cfg: need_reflesh = True
        if not need_reflesh: continue
        replace_item = chain_var.chain_func(*[getattr(default_cfg, v) for v in chain_var.chained_with])
        original_item = getattr(default_cfg, o_key)
        if vb: print靛('[config] warning, %s is chained by %s, automatic modifying:'%(o_key,
                            str(chain_var.chained_with)), original_item, '-->', replace_item)
        setattr(default_cfg, o_key, replace_item)
        altered_cv.append(o_key)
    return altered_cv

def override_config_file(cfg_group, new_cfg, vb):
    import importlib
    assert '->' in cfg_group
    str_pro = '------------- %s -------------'%cfg_group
    if vb:  print绿(str_pro)
    file_, class_ = cfg_group.split('->')
    if '.py' in file_: file_ = file_.rstrip('.py')
    default_configs = getattr(importlib.import_module(file_), class_)
    for key in new_cfg:
        if new_cfg[key] is None: continue
        my_setattr(conf_class=default_configs, key=key, new_value=new_cfg[key], vb=vb)
    altered_cv = secure_chained_vars(default_configs, new_cfg, vb)
    if vb:
        print绿(''.join(['-']*len(str_pro)),)
        arg_summary(default_configs, new_cfg, altered_cv)
        print绿(''.join(['-']*len(str_pro)),'\n\n\n')
    if 'TEAM_NAMES' in new_cfg:
        return [item.split('->')[0] for item in new_cfg['TEAM_NAMES']]
    return None
        
def check_config_relevence(json_data):
    env_name = json_data['config.py->GlobalConfig']['env_name']
    env_path = json_data['config.py->GlobalConfig']['env_path']
    for key in json_data.keys():
        if 'MISSIONS' in key: assert env_path in key, ('configering wrong env!')

def load_config_via_json(json_data, vb):
    for cfg_group in json_data:
        dependency = override_config_file(cfg_group, json_data[cfg_group], vb)
        if dependency is not None:
            for dep in dependency:
                assert any([dep in k for k in json_data.keys()]), 'Arg check failure, There is something missing!'
    check_config_relevence(json_data)
    return None

def get_core_args(vb=True):
    parser = argparse.ArgumentParser(description='HMP')
    parser.add_argument('-c', '--cfg', help='Path of the configuration file')
    parser.add_argument('-s', '--skip', action='store_true', help='skip logdir check')
    args, unknown = parser.parse_known_args()
    load_via_json = (hasattr(args, 'cfg') and args.cfg is not None)
    if load_via_json:
        if len(unknown) > 0 and vb:  
            print亮红('Warning! In json setting mode, %s is ignored'%str(unknown))
        import commentjson as json
        with open(args.cfg) as f: json_data = json.load(f)
        core_group = 'config.py->GlobalConfig'
        override_config_file(core_group, json_data[core_group], vb)
    else:
        assert False
        new_args = load_config_via_cmdline(vb)

    from config import GlobalConfig as cfg
    return cfg



def prepare_args(vb=True):
    parser = argparse.ArgumentParser(description='HMP')
    parser.add_argument('-c', '--cfg', help='Path of the configuration file')
    parser.add_argument('-s', '--skip', action='store_true', help='skip logdir check')
    args, unknown = parser.parse_known_args()
    load_via_json = (hasattr(args, 'cfg') and args.cfg is not None)
    skip_logdir_check = (hasattr(args, 'skip') and (args.skip is not None) and args.skip)
    if load_via_json:
        if len(unknown) > 0 and vb: 
            print亮红('Warning! In json setting mode, %s is ignored'%str(unknown))
        import commentjson as json
        with open(args.cfg) as f:
            json_data = json.load(f)
        new_args = load_config_via_json(json_data, vb)
    else:
        new_args = load_config_via_cmdline(vb)

    from config import GlobalConfig as cfg
    note_name_overide = None
    if not skip_logdir_check: 
        note_name_overide = check_experiment_log_path(cfg.logdir)
        if note_name_overide is not None: 
            override_config_file('config.py->GlobalConfig', {'note':note_name_overide}, vb)
    if not os.path.exists(cfg.logdir): os.makedirs(cfg.logdir)
    if load_via_json and (not cfg.recall_previous_session): 
        copyfile(args.cfg, '%s/experiment.json'%cfg.logdir)
        backup_files(cfg.backup_files, cfg.logdir)
        cfg.machine_info = register_machine_info(cfg.logdir)
    cfg.cfg_ready = True
    return cfg

def register_machine_info(logdir):
    import socket, json, subprocess, uuid
    from .network import get_host_ip
    info = {
        'HostIP': get_host_ip(),
        'ExpUUID':uuid.uuid1().hex,
        'RunPath': os.getcwd(),
        'StartDateTime': time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
    }
    try:
        info['DockerContainerHash'] = subprocess.getoutput(r'cat /proc/self/cgroup | grep -o -e "docker/.*"| head -n 1 |sed "s/docker\\/\\(.*\\)/\\1/" |cut -c1-12')
    except: 
        info['DockerContainerHash'] = 'None'
    with open('%s/info.json'%logdir, 'w+') as f:
        json.dump(info, f, indent=4)
    return info

def backup_files(files, logdir):
    for file in files:
        if os.path.isfile(file):
            print绿('[config] Backup File:',file)
            bkdir = '%s/backup_files/'%logdir
            if not os.path.exists(bkdir): os.makedirs(bkdir)
            copyfile(file, '%s/%s'%(bkdir, os.path.basename(file)))
        else:
            print亮绿('[config] Backup Folder:',file)
            assert os.path.isdir(file), ('cannot find', file)
            copytree(file, '%s/backup_files/%s'%(logdir, os.path.basename(file)), 
                dirs_exist_ok=True, ignore=ignore_patterns("__pycache__"))

    return 

def check_experiment_log_path(logdir):
    res = None
    if os.path.exists(logdir):
        if os.path.exists(logdir+'test_stage'): return None
        print亮红('Current log path:', logdir)
        print亮红('Warning! you will overwrite old logs if continue!')
        print亮红("Pause for 60 seconds before continue (or enter NEW note name!)")
        try:
            res = askChoice()
            if res == '': res = None
        except func_timeout.exceptions.FunctionTimedOut as e:
            res = None
    return res


@func_timeout.func_set_timeout(60)
def askChoice():
    return input('>>')



# def input_or_timeout(timeout):
#     for i in range(30):
#         time.sleep(1)
#     return

def arg_summary(config_class, modify_dict = {}, altered_cv = []):
    for key in config_class.__dict__: 
        if '__' in key: continue
        if key.endswith('_cv'): continue
        if (not key in modify_dict) or (modify_dict[key] is None): 
            if key not in altered_cv: 
                print绿(key.center(25), '-->', str(getattr(config_class,key)))
            else: 
                print靛(key.center(25), '-->', str(getattr(config_class,key)))
        else: 
            print红(key.center(25), '-->', str(getattr(config_class,key)))

def load_config_via_cmdline(vb):
    parser = argparse.ArgumentParser(description='HMP')
    # environment 
    from config import GlobalConfig as cfg
    for setting_name in cfg.__dict__:
        if '__' in setting_name: continue
        if setting_name.endswith('_cv'): continue
        try:
            parser.add_argument('--' + setting_name)
        except:
            if vb: print红('[config] 参数重复！ ', setting_name)
    args = vars(parser.parse_args())
    args = {key: args[key] for key in args if args[key] is not None}
    override_config_file('config.py->GlobalConfig', args, vb)
    if vb: arg_summary(cfg, args)
    return args

def my_setattr(conf_class, key, new_value, vb):
    assert hasattr(conf_class, key), (conf_class, 'has no such config item: **%s**'%key)
    setting_name = key
    replace_item = new_value
    original_item = getattr(conf_class, setting_name)
    if vb: print绿('[config] override %s:'%setting_name, original_item, '-->', replace_item)
    if isinstance(original_item, float):
        replace_item = float(replace_item)
    elif isinstance(original_item, bool):
        if replace_item == 'True':
            replace_item = True
        elif replace_item == 'False':
            replace_item = False
        elif isinstance(replace_item, bool):
            replace_item = replace_item
        else:
            assert False, ('enter True or False, but have:', replace_item)
    elif isinstance(original_item, int):
        replace_item = int(replace_item)
    elif isinstance(original_item, str):
        replace_item = replace_item
    elif isinstance(original_item, list):
        assert isinstance(replace_item, list)
    else:
        assert False, ('not support this type')
    setattr(conf_class, setting_name, replace_item)
    return

def find_all_conf():
    import glob
    py_script_list = glob.glob('./**/*.py', recursive=True)
    conf_class_gather = []
    for python_file in py_script_list:
        with open(python_file,encoding='UTF-8') as f:
            lines = f.readlines()
        for line in lines:
            if 'ADD_TO_CONF_SYSTEM' not in line: continue
            if 'class ' not in line: continue
            conf_class_gather.append({'line':line, 'file':python_file})
    def getBetween(str, str1, str2):
        strOutput = str[str.find(str1)+len(str1):str.find(str2)]
        return strOutput
    for target in conf_class_gather:
        class_name = getBetween(target['line'], 'class ', '(')
        target['class_name'] = class_name
        target['file'] = target['file'].replace('/', '.').replace('..', '')
        import importlib
        target['class'] = getattr(importlib.import_module(target['file'].replace('.py', '')), class_name)
    return conf_class_gather

def make_json(conf_list):
    import json
    out = {}
    for conf in conf_list:
        local_conf = {}
        config_class = conf['class']
        for key in config_class.__dict__: 
            if '__' in key or '_cv' in key: continue
            item_to_be_serialize = getattr(config_class, key)
            try:
                json.dumps(item_to_be_serialize)
            except:
                item_to_be_serialize = '[cannot be json]' + str(item_to_be_serialize)
            local_conf[key] = item_to_be_serialize
        out[conf['file']] = local_conf
    # json_str = json.dumps(out)
    with open('all_conf.json', 'w') as f:
        json.dump(out, f, indent=4)
        print亮紫('the conf summary is successfully saved to all_conf.json')

if __name__ == '__main__':
    conf_list = find_all_conf()
    res_json = make_json(conf_list)


import numpy as np
import time
from MISSIONS.env_router import make_env_function
from UTILS.colorful import print亮红

N = lambda x: np.array(x)

# Here use a pool of multiprocess workers to control a bundle of environment to sync step
# SuperPool.add_target: in each process, initiate a class object named xxxx, 
#     example:
#     self.SuperPool.add_target(name='env', lam=EnvWithRay, args_list=env_args_dict_list)
# SuperPool.exec_target: in each process, make the object (id by name) to call its method
#     example:
#     self.SuperPool.exec_target(name='env', dowhat='step', args_list=actions)
#     self.SuperPool.exec_target(name='env', dowhat='reset')


# ! this class execute in child process
# Ray is much slower compare to our shm/pipe solution,
# we don't use it any more despite the class name
class EnvWithRay(object):
    def __init__(self, env_args_dict):
        env_name = env_args_dict['env_name']
        proc_index = env_args_dict['proc_index']
        env_init_fn = make_env_function(env_name=env_name, rank=proc_index)
        self.env = env_init_fn()
        # finally the env is initialized
        self.observation_space = self.env.observation_space
        self.action_space = self.env.action_space
        self.echo = None

    def step(self, act):
        if np.isnan(act).any(): 
            # env is paused, skip by returning previous obs
            assert self.echo is not None
            return self.echo
        # ! step here
        ob, reward, done, info = self.env.step(act)
        if isinstance(ob, list): 
            print('warning ob is list, which is low-efficient')
            ob = np.array(ob, dtype=object)
        
        if np.any(done):
            # if the environment is terminated, 
            # first, put terminal obs into 'info'
            if info is None:
                info = {'obs-echo':ob}
            else:
                assert isinstance(info, dict), ('oh? info is not dictionary? did not expect that...')
                info.update({'obs-echo': ob.copy()})
            # second, automatically reset env
            ob = self.env.reset()
            if isinstance(ob, tuple):
                # some env like starcraft return (ob, info) tuple at reset
                # have info, then update info
                ob, info_reset = ob
                info = self.dict_update(info, info_reset)
        # preserve an echo here, 
        # will be use to handle unexpected env pause
        self.echo = [ob, reward, done, info]
        # give everything back to main process
        return (ob, reward, done, info)

    def dict_update(self, info, info_reset):
        for key in info_reset:
            if key in info: info[key+'-echo'] = info.pop(key)
        info.update(info_reset)
        return info

    def reset(self):
        return self.env.reset()

    def render(self):
        return self.env.render()

    def close(self):
        return None

    def get_act_space(self):
        return self.action_space

    def get_obs_space(self):
        return self.observation_space

    def get_act_space_str(self):
        return str(self.action_space)

    def get_obs_space_str(self):
        return str(self.observation_space)


# ! this class execute in main process
class SuperpoolEnv(object):
    def __init__(self, process_pool, env_args_dict_list, spaces=None):
        self.SuperPool = process_pool
        self.num_envs = len(env_args_dict_list)
        self.env_name_marker = env_args_dict_list[0][0]['marker']
        self.env = 'env' + self.env_name_marker
        self.SuperPool.add_target(name=self.env, lam=EnvWithRay, args_list=env_args_dict_list)
        try:
            self.observation_space = self.SuperPool.exec_target(name=self.env, dowhat='get_obs_space')[0]
            self.action_space =      self.SuperPool.exec_target(name=self.env, dowhat='get_act_space')[0]
        except:
            print亮红('Gym Space is unable to transfer between processes, using string instead')
            self.observation_space = self.SuperPool.exec_target(name=self.env, dowhat='get_obs_space_str')[0]
            self.action_space =      self.SuperPool.exec_target(name=self.env, dowhat='get_act_space_str')[0]
        # self.observation_space = self.SuperPool.exec_target(name=self.env, dowhat='get_obs_space_str')[0]
        # self.action_space =      self.SuperPool.exec_target(name=self.env, dowhat='get_act_space_str')[0]
        return

    def get_space(self):
        return {'obs_space': self.observation_space, 'act_space': self.action_space}

    def step(self, actions):
        # ENV_PAUSE = [np.isnan(thread_act).any() for thread_act in actions]
        results = self.SuperPool.exec_target(name=self.env, dowhat='step', args_list=actions)
        obs, rews, dones, infos = zip(*results)

        # if any(ENV_PAUSE):
        #     assert not all(ENV_PAUSE)
        #     return self.stack(ENV_PAUSE, obs, rews, dones, infos)
        # else:
        try:
            return np.stack(obs), np.stack(rews), np.stack(dones), np.stack(infos)
        except:
            print('not align! ',obs, rews, dones)
            assert False, ('unalign! ',obs, rews, dones)


    def reset(self):
        results = self.SuperPool.exec_target(name=self.env, dowhat='reset')
        # [ env.reset.remote() for env in self.ray_env_vector])
        if isinstance(results[0], tuple):
            obs, infos = zip(*results)
            return np.stack(obs), np.stack(infos)
        else:
            return np.stack(results)

    # def stack(self, ENV_PAUSE, *args):
    #     sample_index = ENV_PAUSE.index(False)
    #     args = list(args)
    #     for k in range(len(args)):
    #         args[k] = list(args[k])
    #         arg = args[k]
    #         sample = arg[sample_index]
    #         for thread_index in range(self.num_envs):
    #             if arg[thread_index] is not None: continue
    #             if isinstance(sample, np.ndarray):
    #                 arg[thread_index] = np.zeros_like(sample) + np.nan
    #             else:
    #                 assert False
    #         args[k] = np.stack(args[k])
    #     return args


import platform
from sys import stdout

if platform.system()=="Linux":
    pass
else: 
    from colorama import init
    init()

# Do you like the elegance of Chinese characters?
def print红(*kw,**kargs):
    print("\033[0;31m",*kw,"\033[0m",**kargs)
def print绿(*kw,**kargs):
    print("\033[0;32m",*kw,"\033[0m",**kargs)
def print黄(*kw,**kargs):
    print("\033[0;33m",*kw,"\033[0m",**kargs)
def print蓝(*kw,**kargs):
    print("\033[0;34m",*kw,"\033[0m",**kargs)
def print紫(*kw,**kargs):
    print("\033[0;35m",*kw,"\033[0m",**kargs)
def print靛(*kw,**kargs):
    print("\033[0;36m",*kw,"\033[0m",**kargs)

def print亮红(*kw,**kargs):
    print("\033[1;31m",*kw,"\033[0m",**kargs)
def print亮绿(*kw,**kargs):
    print("\033[1;32m",*kw,"\033[0m",**kargs)
def print亮黄(*kw,**kargs):
    print("\033[1;33m",*kw,"\033[0m",**kargs)
def print亮蓝(*kw,**kargs):
    print("\033[1;34m",*kw,"\033[0m",**kargs)
def print亮紫(*kw,**kargs):
    print("\033[1;35m",*kw,"\033[0m",**kargs)
def print亮靛(*kw,**kargs):
    print("\033[1;36m",*kw,"\033[0m",**kargs)

if not stdout.isatty():
    # redirection, avoid a fucked up log file
    print红 = print
    print绿 = print
    print黄 = print
    print蓝 = print
    print紫 = print
    print靛 = print
    print亮红 = print
    print亮绿 = print
    print亮黄 = print
    print亮蓝 = print
    print亮紫 = print
    print亮靛 = print



import setuptools, glob, os

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

    
def _process_requirements():
    packages = open('requirements.txt').read().strip().split('\n')
    requires = []
    for pkg in packages:
        if pkg.startswith('git+ssh'):
            return_code = os.system('pip install {}'.format(pkg))
            assert return_code == 0, 'error, status_code is: {}, exit!'.format(return_code)
        else:
            requires.append(pkg)
    return requires

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

extra_files = package_files('src')



setuptools.setup(
    name="vhmap",
    version="1.6.1",
    author="Qingxu",
    author_email="505030475@qq.com",
    description="Advanced 3D visualizer for researchers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitee.com/hh505030475/hmp-2g/tree/upload-pip/",
    project_urls={
        "Bug Tracker": "https://gitee.com/hh505030475/hmp-2g/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    package_data={"": extra_files},
    include_package_data=True,
    # package_data=[
    #     ('src/VISUALIZE/threejsmod', [f for f in glob.glob('src/VISUALIZE/threejsmod/**/*', recursive=True) if not os.path.isdir(f)]),
    # ],
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=_process_requirements(),
)


import os, copy, atexit, time
import numpy as np
from colorama import init; init()
from multiprocessing import Process
from UTILS.colorful import *
from UTILS.network import get_host_ip, find_free_port

mcom_fn_list_define = [
    "v2dx", "flash", "plot", "figure", "hold", "box", "pause", "clf", "xlim", "ylim", "xlabel", 
    "ylabel", "drawnow", "v2d", "v2d_init", "v3d_init", "v2L", "title", "plot3", "grid", "v3dx", "v2d_show", 
    "v2d_pop", "v2d_line_object", "v2d_clear", "v2d_add_terrain", "set_style", "set_env", "use_geometry", 
    "geometry_rotate_scale_translate", "test_function_terrain", 'line3d', 'advanced_geometry_rotate_scale_translate',
]
别名对齐 = [
    ('初始化3D', 'v2d_init'),
    ('设置样式', 'set_style'),
    ('形状之旋转缩放和平移','geometry_rotate_scale_translate'),
    ('其他几何体之旋转缩放和平移','advanced_geometry_rotate_scale_translate'),
    ('发送几何体','v2dx'),
    ('结束关键帧','v2d_show'),
    ('发送线条','line3d'),
]

# The Design Principle: Under No Circumstance should this program Interrupt the main program!
class mcom():
    def __init__(self, ip=None, port=None, path=None, digit=8, rapid_flush=True, draw_mode=False, tag='default'):
        # digit 默认8，可选4,16，越小程序负担越轻 (all is float, set valid digit number)
        # rapid_flush 当数据流不大时，及时倾倒文件缓存内容 (set 'False' if you'd like your SSD to survive longer)
        self.draw_mode = draw_mode
        self.path = path

        if draw_mode in ['Web', 'Native', 'Img', 'Threejs']:
            self.draw_process = True; print亮红('[mcom.py]: draw process active!')
            port = find_free_port()
            self.draw_tcp_port = ('localhost', port)
            kargs = {
                'draw_mode': draw_mode,
                'draw_udp_port': self.draw_tcp_port,
                'port': self.draw_tcp_port,
                'backup_file': self.path + '/backup.dp'
            }
            DP = DrawProcess if draw_mode != 'Threejs' else DrawProcessThreejs
            self.draw_proc = DP(**kargs)
            self.draw_proc.start()
            self.draw_tcp_client = tcp_client('localhost:%d'%port)
        else:
            print亮红('[mcom.py]: Draw process off! No plot will be done')
            self.draw_process = False

        prev_start, prev_end, self.current_buffer_index = find_free_index(self.path)
        self.starting_file = self.path + '/mcom_buffer_%d____starting_session.txt' % (self.current_buffer_index)
        self.file_lines_cnt = 0
        self.file_max_lines = 300000  # limit file lines to avoid a very large file
        self.digit = digit
        self.rapid_flush = rapid_flush
        self.flow_cnt = 0
        self.tag = tag
        print蓝('[mcom.py]: log file at:' + self.starting_file)

        if not self.draw_mode=='Threejs':
            self.current_file_handle = open(self.starting_file, 'wb+')
        atexit.register(lambda: self.__del__())


    def __del__(self):
        # on the end of the program
        if hasattr(self, 'current_file_handle') and self.current_file_handle is not None:
            end_file_flag = (b'><EndTaskFlag\n')
            self.current_file_handle.write(end_file_flag)
            self.current_file_handle.close()
        if hasattr(self, 'port') and self.port is not None:
            self.disconnect()
        if hasattr(self, 'draw_proc') and self.draw_proc is not None:
            try:
                self.draw_proc.terminate()
                self.draw_proc.join()
            except:
                pass
        print蓝('[mcom.py]: mcom exited! tag: %s'%self.tag)


    def disconnect(self):
        # self.draw_udp_client.close()
        self.draw_tcp_client.close()


    def recall(self, starting_file):
        with open(starting_file,'rb') as f:
            lines = f.readlines()
        r = None
        for l in lines:
            if 'rec_show' in str(l, encoding='utf8'): 
                r = copy.deepcopy(l)
                continue
            self.draw_tcp_client.send_bytes(l)
        if r is not None:
            self.draw_tcp_client.send_bytes(r)
        return None

    '''
        mcom core function: send out/write raw bytes
    '''
    def send(self, data):
        if not self.draw_mode=='Threejs':
            # step 1: add to file
            self.file_lines_cnt += 1
            self.current_file_handle.write(data)
            if self.rapid_flush: self.current_file_handle.flush()
            elif self.flow_cnt>500:
                self.current_file_handle.flush()
                self.flow_cnt = 0
            # step 2: check whether the file is too large, if so move on to next file.
            if self.file_lines_cnt > self.file_max_lines:
                end_file_flag = (b'><EndFileFlag\n')
                self.current_file_handle.write(end_file_flag)
                self.current_file_handle.close()
                self.current_buffer_index += 1
                self.current_file_handle = open((self.path + '/mcom_buffer_%d.txt' % self.current_buffer_index), 'wb+')
                self.file_lines_cnt = 0

        # # step 3: send directive to draw process
        if self.draw_process: 
            # self.draw_udp_client.sendto(data, self.draw_udp_port)
            self.draw_tcp_client.send_bytes(data)

    def rec_init(self, color='k'):
        str_tmp = '>>rec_init(\'%s\')\n' % color
        b_tmp = bytes(str_tmp, encoding='utf8')
        self.send(b_tmp)

    def rec_show(self):
        b_tmp = b'>>rec_show\n'
        self.send(b_tmp)

    def rec_end(self):
        self.send(b'>>rec_end\n')

    def rec_save(self):
        self.send(b'>>rec_save\n')

    def rec_end_hold(self):
        self.send(b'>>rec_end_hold\n')

    def rec_clear(self, name):
        str_tmp = '>>rec_clear("%s")\n' % (name)
        str_tmp = bytes(str_tmp, encoding='utf8')
        self.send(str_tmp)

    def rec(self, value, name):
        value = float(value)
        if self.digit == 16:
            str_tmp = '>>rec(%.16e,"%s")\n' % (value, name)
        elif self.digit == 8:
            str_tmp = '>>rec(%.8e,"%s")\n' % (value, name)
        elif self.digit == 4:
            str_tmp = '>>rec(%.4e,"%s")\n' % (value, name)

        str_tmp = bytes(str_tmp, encoding='utf8')
        self.send(str_tmp)

    def 发送虚幻4数据流(self, x, y, z, pitch, yaw, roll):
        x = float(x)
        y = float(y)
        z = float(z)
        pitch = float(pitch)
        yaw = float(yaw)
        roll = float(roll)
        str_tmp = 'UE4>>(\"agent#1\",%.6e,%.6e,%.6e,%.6e,%.6e,%.6e)\n' % (x, y, z, pitch, yaw, roll)
        str_tmp = bytes(str_tmp, encoding='utf8')
        self.send(str_tmp)

    def 发送虚幻4数据流_多智能体(self, x_, y_, z_, pitch_, yaw_, roll_):
        str_list = ['UE4>>']
        for x, y, z, pitch, yaw, roll in zip(x_, y_, z_, pitch_, yaw_, roll_):
            x = float(x)
            y = float(y)
            z = float(z)
            pitch = float(pitch)
            yaw = float(yaw)
            roll = float(roll)
            str_tmp = '(\"agent#1\",%.5e,%.5e,%.5e,%.5e,%.5e,%.5e)' % (x, y, z, pitch, yaw, roll)
            str_list.append(str_tmp)
            str_list.append(';')
        str_list.append('\n')

        cmd = ''.join(str_list)
        self.send(bytes(cmd, encoding='utf8'))

    def other_cmd(self, func_name, *args, **kargs):
        # func_name = traceback.extract_stack()[-2][2]
        strlist = ['>>', func_name, '(']

        for _i_ in range(len(args)):
            if isinstance(args[_i_], np.ndarray):
                strlist = self._process_ndarray(args[_i_], strlist)
            else:
                strlist = self._process_scalar(args[_i_], strlist)
        if len(kargs)>0:
            for _key_ in kargs:
                if isinstance(kargs[_key_], np.ndarray):
                    strlist = self._process_ndarray(kargs[_key_], strlist, _key_)
                else:
                    strlist = self._process_scalar(kargs[_key_], strlist, _key_)

        if strlist[len(strlist) - 1] == "(": strlist.append(")\n")
        else: strlist[len(strlist) - 1] = ")\n" # 把逗号换成后括号
        self.send(bytes(''.join(strlist), encoding='utf8'))

    def _process_scalar(self, arg, strlist,key=None):
        if key is not None: strlist += '%s='%key
        if isinstance(arg, int):
            strlist.append("%d" % arg)
            strlist.append(",")
        elif isinstance(arg, float):
            if self.digit == 16:  strlist.append("%.16e" % arg)
            elif self.digit == 8: strlist.append("%.8e" % arg)
            elif self.digit == 4: strlist.append("%.4e" % arg)
            strlist.append(",")
        elif isinstance(arg, str):
            strlist.append("\'"); strlist.append(arg)
            strlist.append("\'"); strlist.append(",")
        elif isinstance(arg, list):
            strlist.append(str(arg))
            strlist.append(",")
        elif hasattr(arg, 'dtype') and np.issubdtype(arg.dtype, np.integer):
            strlist.append("%d" % arg)
            strlist.append(",")
        elif hasattr(arg, 'dtype') and np.issubdtype(arg.dtype, np.floating):
            if self.digit == 16:  strlist.append("%.16e" % arg)
            elif self.digit == 8: strlist.append("%.8e" % arg)
            elif self.digit == 4: strlist.append("%.4e" % arg)
            strlist.append(",")
        else:
            print('输入的参数类型不能处理',arg.__class__)
        return strlist

    def _process_ndarray(self, args, strlist, key=None):
        if args.ndim == 1:
            if key is not None: strlist += '%s='%key
            d = len(args)
            sub_list = ["["] + ["%.3e,"%t if (i+1)!=d else "%.3e"%t for i, t in enumerate(args)] + ["]"]
            strlist += sub_list
            strlist.append(",")
        elif args.ndim == 2:
            print红('mcom：输入数组的维度大于1维，目前处理不了。')
        else:
            print红('mcom：输入数组的维度大于2维，目前处理不了。')
        return strlist

    for fn_name in mcom_fn_list_define:
        build_exec_cmd = 'def %s(self,*args,**kargs):\n self.other_cmd("%s", *args,**kargs)\n'%(fn_name, fn_name)
        exec(build_exec_cmd)

    for 别名, fn_name in 别名对齐:
        build_exec_cmd = '%s = %s\n'%(别名, fn_name)
        exec(build_exec_cmd)

def find_free_index(path):
    if not os.path.exists(path): os.makedirs(path)

    def find_previous_start_end():
        start = None; end = None; t = 0
        while True:
            is_body = os.path.exists(path + '/mcom_buffer_%d.txt' % t)
            is_head = os.path.exists(path + '/mcom_buffer_%d____starting_session.txt' % t)
            if is_head: start = t
            if is_head or is_body: end = t; t += 1
            else:
                new = t
                return (start, end, new)

    prev_start, prev_end, new = find_previous_start_end()
    return prev_start, prev_end, new

class tcp_client():
    def __init__(self, ip):
        import socket
        TCP_IP, TCP_PORT = ip.split(':')
        TCP_PORT = int(TCP_PORT)
        BUFFER_SIZE = 10240
        self.socketx = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socketx.connect((TCP_IP, TCP_PORT))

        self.fixed_end = bytes('@end@', encoding='utf8') 

    def send(self, str_msg):
        msg = str_msg + '@end@'
        self.socketx.send(bytes(msg, encoding='utf8'))

    def send_bytes(self, b_msg):
        msg = b_msg + self.fixed_end
        self.socketx.send(msg)

    def close(self):
        self.socketx.close()

    def __del__(self):
        self.socketx.close()

class tcp_server():
    def __init__(self, ip_port):
        import socket
        self.draw_cmd_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.draw_cmd_socket.bind(ip_port)
        self.draw_cmd_socket.listen()
        self.handler = None
        self.queue = None
        self.buff = ['']
 
    def wait_connection(self):
        import threading
        self.sock, _ = self.draw_cmd_socket.accept()
        t = threading.Thread(target=self.listening_thread)
        t.daemon = True
        t.start()

    def listening_thread(self):

        def handle_flag_breakdown():
            split_ = self.buff[-1].split('@end@')
            assert len(split_)==2
            self.buff[-1] = split_[0]
            self.buff.append('')
            self.buff[-1] = split_[1]
            return

        assert (self.handler is None)     or (self.queue is None)
        assert (self.handler is not None) or (self.queue is not None)

        while True:
            recvData = self.sock.recv(10240)
            recvData = str(recvData, encoding = "utf-8")
            ends_with_mark = recvData.endswith('@end@')
            split_res = recvData.split('@end@')
            assert len(split_res) != 0
            if len(split_res) == 1:
                # 说明没有终止符，直接将结果贴到buf最后一项
                self.buff[-1] = self.buff[-1] + split_res[0]
                if '@end@' in self.buff[-1]: 
                    handle_flag_breakdown()
            else:
                n_split = len(split_res)
                for i, r in enumerate(split_res):
                    self.buff[-1] = self.buff[-1] + r
                    if i == 0 and '@end@' in self.buff[-1]:
                        handle_flag_breakdown()
                    if i != n_split-1:
                        self.buff.append('')
                    else: # 最后一个
                        if r == '': continue
            buff_list = self.buff[:-1]
            self.buff = self.buff[-1:]
            if self.handler is not None: 
                self.handler(buff_list)
            if self.queue is not None: 
                self.queue.put(buff_list)
        return

    def set_handler(self, handler):
        self.handler = handler

    def get_queue(self):
        import queue
        self.queue = queue.Queue()
        return self.queue

    def recv(self):
        return

    # def send(self, str_msg):
    #     msg = '@start@' + str_msg + '@end@'
    #     self.sock.send(bytes(msg, encoding='utf8'))

    def send_bytes(self, b_msg):
        msg = self.fixed_L + b_msg + self.fixed_R
        self.sock.send(msg)

    def close(self):
        if hasattr(self, 'sock'): self.sock.close()
        if hasattr(self, 'draw_cmd_socket'): self.draw_cmd_socket.close()        

    def __del__(self):
        self.close()





class MyHttp(Process):
    def __init__(self, path_to_html, avail_port):
        super(MyHttp, self).__init__()
        self.path_to_html = path_to_html
        self.avail_port = avail_port

    def run(self):
        from flask import Flask
        app = Flask(__name__)
        @app.route("/")
        def hello():
            try:
                with open(self.path_to_html,'r') as f:
                    html = f.read()
            except:
                html = "no plot yet please wait"
            return html
        app.run(port=self.avail_port)



class DrawProcessThreejs(Process):
    def __init__(self, draw_udp_port, draw_mode, **kargs):
        super(DrawProcessThreejs, self).__init__()
        self.draw_mode = draw_mode
        self.draw_udp_port = draw_udp_port
        self.tcp_connection = tcp_server(self.draw_udp_port)
        self.buffer_list = []
        self.backup_file = kargs['backup_file']
        self.allow_backup = False if self.backup_file is None else True
        self.flask_thread = None
        self.waitress_server = None

        if self.allow_backup:
            if os.path.exists(self.backup_file):
                print亮红('[mcom.py]: warning, purge previous 3D visual data!')
                os.remove(self.backup_file)

    def init_threejs(self):
        import threading
        self.flask_thread = threading.Thread(target=self.run_flask, args=(find_free_port(),))
        self.flask_thread.daemon = True
        self.flask_thread.start()

    def run(self):
        self.init_threejs()
        try:
            from queue import Empty
            queue = self.tcp_connection.get_queue()
            self.tcp_connection.wait_connection() # after this, the queue begin to work
            while True:
                self.run_handler(queue.get(timeout=600))
        except KeyboardInterrupt:
            self.__del__()
        self.__del__()

    def __del__(self):
        if self.waitress_server is not None:
            self.waitress_server.close()
            print('self.waitress_server.close()')
        return
        
    def run_handler(self, new_buff_list):
        self.buffer_list.extend(new_buff_list)
        if self.allow_backup:
            with open(self.backup_file, 'a+') as f: f.writelines(new_buff_list)
        # too many, delete with fifo
        if len(self.buffer_list) > 1e9: # 当存储的指令超过十亿后，开始删除旧的
            del self.buffer_list[:len(new_buff_list)]

    def run_flask(self, port):
        from flask import Flask, url_for, jsonify, request, send_from_directory, redirect
        from waitress import serve

        app = Flask(__name__)
        dirname = os.path.dirname(__file__) + '/threejsmod'
        import zlib

        @app.route("/up", methods=["POST"])
        def upvote():
            # dont send too much in one POST, might overload the network traffic
            if len(self.buffer_list)>35000:
                tosend = self.buffer_list[:30000]
                self.buffer_list = self.buffer_list[30000:]
            else:
                tosend = self.buffer_list
                self.buffer_list = []

            # use zlib to compress output command, worked out like magic
            buf = "".join(tosend)
            buf = bytes(buf, encoding='utf8')   
            zlib_compress = zlib.compressobj()
            buf = zlib_compress.compress(buf) + zlib_compress.flush(zlib.Z_FINISH)
            return buf

        @app.route("/<path:path>")
        def static_dirx(path):
            if path=='favicon.ico': 
                return send_from_directory("%s/"%dirname, 'files/HMP.ico')
            return send_from_directory("%s/"%dirname, path)

        @app.route("/")
        def main_app():
            with open('%s/examples/abc.html'%dirname, 'r', encoding = "utf-8") as f:
                buf = f.read()
            return buf

        print('\n--------------------------------')
        print('JS visualizer online: http://%s:%d'%(get_host_ip(), port))
        print('JS visualizer online (localhost): http://localhost:%d'%(port))
        print('--------------------------------')

        def serve(app, **kw):
            import logging
            from waitress.server import create_server
            _server = kw.pop("_server", create_server)  # test shim
            _quiet = kw.pop("_quiet", False)  # test shim
            _profile = kw.pop("_profile", False)  # test shim
            if not _quiet:  # pragma: no cover
                # idempotent if logging has already been set up
                logging.basicConfig()
            self.waitress_server = _server(app, **kw)
            if not _quiet:  # pragma: no cover
                self.waitress_server.print_listen("Serving on http://{}:{}")
            # if _profile:  # pragma: no cover
            #     profile("server.run()", globals(), locals(), (), False)
            else:
                self.waitress_server.run()

        serve(app, threads=8, ipv4=True, ipv6=True, listen='*:%d'%port)


class DrawProcess(Process):
    def __init__(self, draw_udp_port, draw_mode, **kargs):
        super(DrawProcess, self).__init__()
        self.draw_mode = draw_mode
        self.draw_udp_port = draw_udp_port
        self.tcp_connection = tcp_server(self.draw_udp_port)

        return

    def init_matplot_lib(self):
        if self.draw_mode in ['Web', 'Img']:
            import matplotlib
            matplotlib.use('Agg') # set the backend before importing pyplot
            import matplotlib.pyplot as plt
            self.gui_reflesh = lambda: time.sleep(1) # plt.pause(0.1)
        elif self.draw_mode == 'Native':
            import matplotlib
            # matplotlib.use('Agg') # set the backend before importing pyplot
            matplotlib.use('Qt5Agg')
            import matplotlib.pyplot as plt
            self.gui_reflesh = lambda: plt.pause(0.2) #canvas.start_event_loop(0.1) # plt.pause(0.2) #time.sleep(0.2) #
        elif self.draw_mode == 'Threejs':
            assert False
        else:
            assert False

        from config import GlobalConfig
        logdir = GlobalConfig.logdir
        if not os.path.exists(logdir):
            os.makedirs(logdir)
        if self.draw_mode == 'Web':
            self.avail_port = find_free_port()
            my_http = MyHttp('%s/html.html'%logdir, self.avail_port)
            my_http.daemon = True
            my_http.start()
        self.libs_family = {
            'rec_init': 'rec', 'rec': 'rec', 'rec_show': 'rec',
            'v2d_init': 'v2d', 'v2dx':'v2d', 'v2d_show': 'v2d', 'v2d_pop':'v2d',
            'v2d_line_object':'v2d', 'v2d_clear':'v2d', 'v2d_add_terrain': 'v2d',
        }
        self.libs_init_fns = {
            'rec': self.rec_init_fn,
            'v2d': self.v2d_init_fn,
        }

    def run(self):
        self.init_matplot_lib()
        try:
            # self.tcp_connection.set_handler(self.run_handler)
            from queue import Empty
            queue = self.tcp_connection.get_queue()
            # self.tcp_connection.set_handler(self.run_handler)
            self.tcp_connection.wait_connection() # after this, the queue begin to work
            while True:
                try: self.run_handler(queue.get(timeout=0.1))
                except Empty: self.gui_reflesh()

        except KeyboardInterrupt:
            self.__del__()
        self.__del__()

    def run_handler(self, buff_list):
        for buff in buff_list:
            self.process_cmd(buff)
            # print('成功处理指令:', buff)

    def __del__(self):
        self.tcp_connection.close()



    def process_cmd(self, cmd_str):
        if '>>' in cmd_str:
            cmd_str_ = cmd_str[2:].strip('\n')
            if ')' not in cmd_str_:
                cmd_str_ = cmd_str_+'()'
            prefix = self.get_cmd_lib(cmd_str_)
            if prefix is not None: 
                eval('%s.%s'%(prefix, cmd_str_))

    def get_cmd_lib(self, cmd):
        cmd_key = None
        func_name = cmd.split('(')[0]
        if func_name not in self.libs_family:
            print蓝('绘图函数不能处理：', cmd)
            return None
        family_name = self.libs_family[func_name]
        if self.libs_init_fns[family_name] is not None:
            self.libs_init_fns[family_name]()
            self.libs_init_fns[family_name] = None
        return 'self.%s'%family_name

    def rec_init_fn(self):
        from VISUALIZE.mcom_rec import rec_family
        self.rec = rec_family('r', self.draw_mode)

    def v2d_init_fn(self):
        from VISUALIZE.mcom_v2d import v2d_family
        self.v2d = v2d_family(self.draw_mode)




"""
Created on Tue Aug 22 19:41:55 2017
@author: Quantum Liu
"""
'''
Example:
gm=GPUManager()
with torch.cuda.device(gm.auto_choice()):
    blabla
Or:
gm=GPUManager()
torch.cuda.set_device(gm.auto_choice())
'''

import os, time
from UTILS.colorful import print黄

class sel_gpu():
    '''
    qargs:
        query arguments
    A manager which can list all available GPU devices
    and sort them and choice the most free one.Unspecified 
    ones pref.
    GPU设备管理器，考虑列举出所有可用GPU设备，并加以排序，自动选出
    最空闲的设备。在一个GPUManager对象内会记录每个GPU是否已被指定，
    优先选择未指定的GPU。
    '''
    def __init__(self,qargs=[]):
        '''
        '''
        
        self.qargs=qargs


    def _sort_by_memory(self,gpus,by_size=False):
        for gpu in gpus:    # 优先使用A100显卡
            if 'A100' in gpu['gpu_name']:
                gpu['memory.free'] *= 2
                gpu['memory.total'] *= 2

        if by_size:
            print黄('Sorted by free memory size')
            # for gpu in gpus:    # 优先使用A100显卡
            #     if 'A100' in gpu['gpu_name']:
            #         gpu['memory.free'] *= 2
            #         gpu['memory.total'] *= 2
            res = sorted(gpus,key=lambda d:d['memory.free'],reverse=True)
            # for gpu in gpus:    # 优先使用A100显卡
            #     if 'A100' in gpu['gpu_name']:
            #         gpu['memory.free'] /= 2
            #         gpu['memory.total'] /= 2

            return res
        else:
            print黄('Sorted by free memory rate')
            return sorted(gpus,key=lambda d:float(d['memory.free'])/ d['memory.total'],reverse=True)


    def _sort_by_power(self,gpus):
        return sorted(gpus,key='by_power')
    
    def _sort_by_custom(self,gpus,key,reverse=False,qargs=[]):
        if isinstance(key,str) and (key in qargs):
            return sorted(gpus,key=lambda d:d[key],reverse=reverse)
        if isinstance(key,type(lambda a:a)):
            return sorted(gpus,key=key,reverse=reverse)
        raise ValueError("The argument 'key' must be a function or a key in query args,please read the documention of nvidia-smi")

    def auto_choice(self,mode=0):
        '''
        mode:
            0:(default)sorted by free memory size
        return:
            a TF device object
        Auto choice the freest GPU device,not specified
        ones 
        自动选择最空闲GPU,返回索引
        '''
        from UTILS.colorful import print黄
        self.gpus=self.query_gpu(self.qargs)
        for gpu in self.gpus:
            gpu['specified']=False
        self.gpu_num=len(self.gpus)
        # if not self.check_gpus():
        #     raise ImportError('GPU available check failed')

        for old_infos,new_infos in zip(self.gpus,self.query_gpu(self.qargs)):
            old_infos.update(new_infos)
        unspecified_gpus=[gpu for gpu in self.gpus if not gpu['specified']] or self.gpus
        
        if mode==0:
            chosen_gpu=self._sort_by_memory(unspecified_gpus,True)[0]
            print黄('Choosing the GPU device has largest free memory...\n')
        elif mode==1:
            chosen_gpu=self._sort_by_power(unspecified_gpus)[0]
            print黄('Choosing the GPU device has highest free memory rate...\n')
        elif mode==2:
            chosen_gpu=self._sort_by_power(unspecified_gpus)[0]
            print黄('Choosing the GPU device by power...\n')
        else:
            chosen_gpu=self._sort_by_memory(unspecified_gpus)[0]
            print黄('Given an unaviliable mode,will be chosen by memory\n')
        chosen_gpu['specified']=True
        index=chosen_gpu['index']
        print黄('Using GPU {i}:\n{info}'.format(i=index,info='\n'.join([str(k)+':'+str(v) for k,v in chosen_gpu.items()])))
        return int(index)


    @staticmethod
    def check_gpus():
        '''
        GPU available check
        http://pytorch-cn.readthedocs.io/zh/latest/package_references/torch-cuda/
        '''
        import torch
        if not torch.cuda.is_available():
            print黄('This script could only be used to manage NVIDIA GPUs,but no GPU found in your device')
            return False


        with os.popen('nvidia-smi -h') as f:
            if not 'NVIDIA System Management' in f.read():
                print黄("'nvidia-smi' tool not found.")
                f.close()
                return False
            f.close()
        return True

    @staticmethod
    def parse(line,qargs):
        '''
        line:
            a line of text
        qargs:
            query arguments
        return:
            a dict of gpu infos
        Pasing a line of csv format text returned by nvidia-smi
        解析一行nvidia-smi返回的csv格式文本
        '''
        numberic_args = ['memory.free', 'memory.total', 'power.draw', 'power.limit']#可计数的参数
        power_manage_enable=lambda v:(not 'Not Support' in v)#lambda表达式，显卡是否滋瓷power management（笔记本可能不滋瓷）
        to_numberic=lambda v:float(v.upper().strip().replace('MIB','').replace('W',''))#带单位字符串去掉单位
        process = lambda k,v:((int(to_numberic(v)) if power_manage_enable(v) else 1) if k in numberic_args else v.strip())
        return {k:process(k,v) for k,v in zip(qargs,line.strip().split(','))}

    def query_gpu(self, qargs=[]):
        '''
        qargs:
            query arguments
        return:
            a list of dict
        Querying GPUs infos
        查询GPU信息
        '''
        qargs =['index','gpu_name', 'memory.free', 'memory.total', 'power.draw', 'power.limit']+ qargs
        cmd = 'nvidia-smi --query-gpu={} --format=csv,noheader'.format(','.join(qargs))
        results = os.popen(cmd).readlines()
        return [self.parse(line,qargs) for line in results]


    @staticmethod
    def by_power(d):
        '''
        helper function fo sorting gpus by power
        '''
        power_infos=(d['power.draw'],d['power.limit'])
        if any(v==1 for v in power_infos):
            print黄('Power management unable for GPU {}'.format(d['index']))
            return 1
        return float(d['power.draw'])/d['power.limit']



#coding=utf-8
import glob,os,sys,re,subprocess,platform


def print红(*kw):
    print("\033[0;31m",*kw,"\033[0m")
def print绿(*kw):
    print("\033[0;32m",*kw,"\033[0m")
def print黄(*kw):
    print("\033[0;33m",*kw,"\033[0m")
def print蓝(*kw):
    print("\033[0;34m",*kw,"\033[0m")
def print紫(*kw):
    print("\033[0;35m",*kw,"\033[0m")
def print靛(*kw):
    print("\033[0;36m",*kw,"\033[0m")
def printX(*kw):
    print("\033[0;38m",*kw,"\033[0m")

# 用pip执行安装指令
def install(package):
    try:
        subprocess.check_call([sys.executable, "-m", "pip", "install",
        "-i","https://pypi.tuna.tsinghua.edu.cn/simple",
        "--progress-bar","emoji",
        "--prefer-binary", 
        package])
    except:
        print红("执行命令 ", "pip", "install","-i","https://pypi.tuna.tsinghua.edu.cn/simple", package, "时，抛出错误")
        pass

sys_name = platform.system()
if sys_name == "Windows":
    try:
        from colorama import init,Fore,Back,Style
        init(autoreset=False)
        def print红(*kw):
            print(Fore.RED,*kw)
        def print绿(*kw):
            print(Fore.GREEN,*kw)
        def print黄(*kw):
            print(Fore.YELLOW,*kw)
        def print蓝(*kw):
            print(Fore.BLUE,*kw)
        def print紫(*kw):
            print(Fore.MAGENTA,*kw)
        def print靛(*kw):
            print(Fore.CYAN,*kw)

    except:
        install('colorama')
        print('颜色组件安装完成！现在请重新运行！')
        sys_name.exit(0)

"""
# step 1, 查询所有子路径.py脚本文件，列表
"""
py_script_list = glob.glob('./**/*.py', recursive=True)
required = []
local_name_list = {"None":False}
引发连锁错误的包_列表 = {"None":False}


"""
# step 2, 提取 import 以及 from *** import
"""
def 是否为工程内的文件交叉调用(包,python_file):
    包_org = 包
    if '.' not in 包:
        res = os.path.exists("./"+包+".py")
        if res:
            return True,包_org
        else:
            return False, 包_org
    if 包.startswith('.'):
        包 = os.path.dirname(python_file).replace("/", ".").replace("..", ".") + 包
        包_org = 包
    包 = 包.replace(".", "/")
    res = os.path.exists("./"+包+".py")
    if res:
        tmp = 包_org.split(".")
        if tmp[0]!='': local_name_list[tmp[0]] = True
        return True, 包_org
    else:
        return False, 包_org

for python_file in py_script_list:
    with open(python_file,encoding='UTF-8') as f:
        lines = f.readlines()
        for line in lines:
            if "import" in line or "from" in line:
                t = line.split()
                # from 开头 或者 import开头
                if t[0] == "import" or t[0] == "from":
                    i = 1
                    包 = ""
                    for ti in t[1:]:
                        if (ti!="import") and (ti!="as"):
                            包 = 包 + ti
                        else:
                            break
                    if "," in 包:
                        包_l = 包.split(",")
                    else:
                        包_l = [包]
                    for 包 in 包_l:
                        包_debug = 包
                        if 包_debug == '.':
                            continue
                        res,包 = 是否为工程内的文件交叉调用(包,python_file)
                        if not res:
                            required.append(包)
                            


required = set(required)
required = sorted(required)

"""
# step 3, 尝试import，筛查缺失的包
"""
print黄("**************************************************************")
print黄("尝试import")
# 使用清华镜像
need_fix_cmd_orig = "pip install -i https://pypi.tuna.tsinghua.edu.cn/simple "
need_fix_cmd      = "pip install -i https://pypi.tuna.tsinghua.edu.cn/simple "
need_fix_list = []
failed_cmd = []
chain_failed = []
for 包 in required:
    cmd = "import "+包
    try:
        # 如果这里罕见地报错，
        # 说明该文件有 import开头的、被“”“包裹的注释, 
        # 找到它，然后删除这个奇葩注释
        exec(cmd)
    except ImportError as error:
        print红("error trying to do:",cmd,error.msg)
        error_str = error.msg.split('\'')
        package_import_error = (len(error_str) >= 2)
        if not package_import_error:
            continue
        包_error = error_str[1]
        if '.' in 包:
            包_l = 包.split('.')
            包_tmp = 包_l[0]
        # 引发问题的不是这个包本身，而是这个包import其他包，但这个包内引包失败了
        # 仅仅是一个连锁错误而已，无需处理
        if 包_error != 包:
            print红("发生连锁引包错误: ",error.msg)
            chain_failed.append(error.msg)
            cmd = cmd + "\t\t此项仅仅由连锁引包错误导致: " + error.msg
            if 包_tmp not in 引发连锁错误的包_列表:
                引发连锁错误的包_列表[包_tmp]=True
        else:
            # 非连锁，一定是真的缺
            引发连锁错误的包_列表[包]=False
        failed_cmd.append(cmd)
        if '.' in 包:
            包_l = 包.split('.')
            包 = 包_l[0]
        if len(包)>19: # some comment mixed in somehow
            continue
        need_fix_list.append(包)
    except BaseException as error:
        print红(error)
    else:
        print绿("this package is ok:",cmd)

need_fix_list = set(need_fix_list)
need_fix_list = sorted(need_fix_list)
if len(failed_cmd) > 0:
    print红("以下的包import操作失败")
    for cmd in failed_cmd:
        print红(cmd)


"""
# step 4, 处理缺失的包，并找到对应的pip安装指令
"""
term_replace_dict = {
    "cv2":"opencv-python",
    "torch":"torch",
    "mpi":"mpi4py",
    "MPI":"mpi4py",
    "mujoco_py":"None",    # pip cannot install this????
    "pybullet_envs":"None",
    "stable_baselines3":"None",
    "pyximport":"cython",
    "PIL":"None",
    "collective_assult":"None",
    "gym_fortattack":"None",
    "multiagent":"None",
    "z_config":"None",
    "gym_vecenv":"None"
}

PIL

for inx, 包 in enumerate(need_fix_list):
    if 包 in term_replace_dict:
        包 = term_replace_dict[包]
        need_fix_list[inx] = 包
    if (包 in local_name_list) or (包 in 引发连锁错误的包_列表 and 引发连锁错误的包_列表[包]==True):
        need_fix_list[inx] = "None"

need_fix_list = set(need_fix_list)
need_fix_list = sorted(need_fix_list)
if len(need_fix_list) == 0:
    print绿("所有依赖已就绪")
    exit(0)

"""
# step 5, 如果有requirement.txt，从中提取出有用的版本信息
"""
print黄("**************************************************************")
print蓝("requirement.txt中的相关信息")
execute_fix = []
if os.path.exists("./requirements.txt"):
    with open("./requirements.txt",encoding='UTF-8') as f:
        lines = f.readlines()
        for line in lines:
            if line.startswith("-"):
                print蓝("requirement.txt要求以下版本： -->   "+"pip install "+line[:-1])
                print蓝("首先git clone，然后找到setup.py的路径，然后执行 pip install --no-deps  -e .")
                continue
            line_split = line.split("==")
            if (len(line_split)==2) and (line_split[0] in need_fix_list):
                print蓝("requirement.txt要求以下版本： -->   "+"pip install "+line[:-1])



"""
# step 6, 如果需要安装pytorch，gym等特殊包，对应给出安装建议
"""

def config_anaconda():
    with open(__file__,'r') as f:
        conda_cmd = f.readlines()
        condarc_lines = conda_cmd[-18:-2]
    f = open('./.condarc','w+')
    f.writelines(condarc_lines)
    f.close()


print黄("**************************************************************")
try:
    conda_env_name = sys.executable.split('/')[-3]
except:
    conda_env_name = sys.executable.split('\\')[-2]
for 包 in need_fix_list:
    if 包 == "torch":
        print蓝("pytorch需要手动安装，pytorch 的安装方法（选择其一），然后重新运行该脚本:")
        print蓝("conda install -n %s  pytorch torchvision torchaudio cudatoolkit=10.2 -c https://mirrors.tuna.tsinghua.edu.cn/anaconda/cloud/pytorch/"%conda_env_name)
        print蓝("conda install -n %s  pytorch torchvision torchaudio cudatoolkit=11.0 -c https://mirrors.tuna.tsinghua.edu.cn/anaconda/cloud/pytorch/"%conda_env_name)
        print蓝("")
        
        sys.exit(0)
    # if 包 == "tensorflow":
    #     print靛("Tensorflow需要手动安装，首先，更换conda源的指令")
    #     config_anaconda()
    #     print靛("cp",os.getcwd()+"/.condarc","~/.condarc")
    #     print靛("然后，安装TF一代的指令")
    #     print靛("conda install -n %s tensorflow-gpu=1.*"%conda_env_name)
    #     sys.exit(0)
    if (包 is not "None"):
        need_fix_cmd = need_fix_cmd + 包 + "  "
        execute_fix.append(包)
print黄("**************************************************************")
print绿(need_fix_cmd)
print黄("**************************************************************")




"""
# step 7, 对于除了特殊包之外的其他软件包，调用pip直接安装
"""
print绿("注意！当前的conda环境是：",conda_env_name," 所有操作都将只在该conda环境内生效")
input("执行自动安装？")
if input("确定执行自动安装？(y/n)")=='y':
    for 包 in execute_fix:
        install(包)


"""
# step 8, 完成任务，取消以下代码的注释，测试pytorch是否工作
"""
        
# import torch
# flag = torch.cuda.is_available()
# print(flag)

# ngpu= 1
# # Decide which device we want to run on
# device = torch.device("cuda:0" if (torch.cuda.is_available() and ngpu > 0) else "cpu")
# print(device)
# print(torch.cuda.get_device_name(0))
# print(torch.rand(3,3).cuda()) 

'''
不要修改或者删除以下内容！！有用！！
channels:
  - defaults
show_channel_urls: true
channel_alias: https://mirrors.tuna.tsinghua.edu.cn/anaconda
default_channels:
  - https://mirrors.tuna.tsinghua.edu.cn/anaconda/pkgs/main
  - https://mirrors.tuna.tsinghua.edu.cn/anaconda/pkgs/free
  - https://mirrors.tuna.tsinghua.edu.cn/anaconda/pkgs/r
  - https://mirrors.tuna.tsinghua.edu.cn/anaconda/pkgs/pro
  - https://mirrors.tuna.tsinghua.edu.cn/anaconda/pkgs/msys2
custom_channels:
  conda-forge: https://mirrors.tuna.tsinghua.edu.cn/anaconda/cloud
  msys2: https://mirrors.tuna.tsinghua.edu.cn/anaconda/cloud
  bioconda: https://mirrors.tuna.tsinghua.edu.cn/anaconda/cloud
  menpo: https://mirrors.tuna.tsinghua.edu.cn/anaconda/cloud
  pytorch: https://mirrors.tuna.tsinghua.edu.cn/anaconda/cloud
  simpleitk: https://mirrors.tuna.tsinghua.edu.cn/anaconda/cloud
'''